#ifndef __chips_BUS_H
#define __chips_BUS_H
/*
 * Copyright (c) 1996-1998 Picture Elements, Inc.
 *    Stephen Williams <steve@picturel.com>
 *
 *    This source code is free software; you can redistribute it
 *    and/or modify it in source code form under the terms of the GNU
 *    Library General Public License as published by the Free Software
 *    Foundation; either version 2 of the License, or (at your option)
 *    any later version. In order to redistribute the software in
 *    binary form, you will need a Picture Elements Binary Software
 *    License.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU Library General Public License for more details.
 *
 *    You should have received a copy of the GNU Library General Public
 *    License along with this program; if not, write to the Free
 *    Software Foundation, Inc.,
 *    59 Temple Place - Suite 330
 *    Boston, MA 02111-1307, USA
 *
 *    You should also have recieved a copy of the Picture Elements
 *    Binary Software License offer along with the source. This offer
 *    allows you to obtain the right to redistribute the software in
 *    binary (compiled) form. If you have not received it, contact
 *    Picture Elements, Inc.,
 *    777 Panoramic Way
 *    Berkeley, CA 94704.
 */
#ident "$Id: BUS.h,v 1.29 1998/06/10 16:53:49 steve Exp $"

# include  <stddef.h>
# include  <ucr/ILock.h>
# include  <ucr/RingBuffer.h>

/*
 * This class implements that uCR/host communication protocol for
 * chatting with the host. The BUS class manages the protocol handling
 * for a set of channels that can be opened seperately. The programmer
 * creates a Device object to represent the physical BUS interface, and
 * Channel objects to represent each interesting channel. The BUS
 * class itself only has public methods for attaching these other
 * objects to it.
 *
 * Details of the BUS protocol tables are in the dicomentation
 * included with this source code.
 *
 * The BUS object does not buffer the data. It is up to the
 * application to perform reasonable I/O operations, and the BUS class
 * will not transfer any bytes until explicitly requested.
 *
 * Setup:
 * - create an instance of the BUS::Device class. This device is
 * used to interface with the hardware.
 *
 * - Create an instance of BUS to manage the physical bus and use
 * open_device to attach the device to the BUS.
 *
 * - Create BUS::Channel objects however you see fit. Each channel
 * must be attached to the bus with open_channel.
 */

class BUS  {

    public:

      class Channel;
      class Frame;
      class Device;

    public:
	// The Channel class represents a single open channel to the
	// host. This class has methods for reading from/writing to
	// the channel that it represents.
      class Channel {

	    friend class BUS;

	  public:
	    explicit Channel(unsigned channel);
	    ~Channel();

	    unsigned channel_id() const { return channel_; }

	    bool read_eof() const { return read_eof_; }
	    void clear_eof()      { read_eof_ = false; }

	      // Return true if at least one byte can be read from the
	      // channel. Generally, this should be protected from
	      // interrupts so that it returns accurate results.
	    bool readable() const;

	    typedef void (*notify_fun)();
	    void readable(notify_fun f);


	      // read the number of bytes from the channel into the
	      // buffer. This method blocks until all the requested
	      // bytes are transferred. If the pointer is null, the
	      // data will be read from the channel, but discarded.

	    size_t read(void*, size_t bytes, bool partial_flag =false);
	    void write(const void*buffer, size_t count);


	      // Set a function to receive attentions on the channel.

	    void attention(void (*fun)(void*, unsigned short, unsigned long), void*arg)
	    { atten_fun_ = fun; atten_arg_ = arg; }

	  private:

	    BUS* master_;
	    const unsigned channel_;
	    Channel*list_; // The master_ uses this.

	      // Some channel state flags
	    bool read_eof_;

	      // Keep information about the channel table on the
	      // host. The BUS class manages these for me, I provide
	      // channel specific space.
	    mutable unsigned long table_;
	    mutable unsigned long magic_;

	      // Information about the buffer I'm currently reading
	      // from. The host sends these to m through the channel
	      // table.
	    mutable unsigned long raddr_;
	    mutable unsigned long rsize_;
	    mutable unsigned long roff_;
	    static bool wait_for_raddr_(const volatile void*);

	    mutable unsigned long waddr_;
	    mutable unsigned long wsize_;
	    mutable unsigned long woff_;
	    static bool wait_for_waddr_(const volatile void*);

	      // These members handle the dispatch of attention
	      // messages from the host to the channel. If the fun_ is
	      // not set, the attention signals are ignored.

	    void (*atten_fun_)(void*, unsigned short, unsigned long);
	    void* atten_arg_;

	    void do_attention_(unsigned short arg1, unsigned long arg2);

	    notify_fun readable_fun_;

	  private: // Not implemented
	    Channel(const Channel&);
	    Channel& operator= (const Channel&);
      };

      void open_channel(Channel*);
      void close_channel(Channel*);

    public:

	/* The class Device represents a physical interface to the BUS
	   bus. Although this is currently coded to use the PLX 9060
	   chip, this class will eventually be made abstract so that
	   many different interfaces can be supported. */

      class Device {

	    friend class BUS;
	    friend class BUS::Channel;

	  public:
	    explicit Device() { }
	    virtual ~Device() =0;

	    virtual unsigned long readl(unsigned long addr) =0;
	    virtual void writel(unsigned long addr, unsigned long val) =0;

	  private:
	      // The BUS object uses these methods to make the device
	      // do the things it wants done.

	    void bus_dma_read(unsigned long bus,
			      void*buffer,
			      size_t count)
	    {
		  bus_transfer(bus, buffer, count, false);
		  bus_transfer_done();
	    }

	    void bus_dma_write(unsigned long bus,
			       const void*buffer,
			       size_t count)
	    {
		  bus_transfer(bus, (void*)buffer, count, true);
		  bus_transfer_done();
	    }

	      // When the device is attached to the BUS, the BUS
	      // object calls this method to start things going.

	    void device_init(BUS*);


	      // The derived class uses this method to deliver the mask
	      // of bell interrupts to this base.
	  protected:
	    void bus_bell_isr(unsigned long mask);

	      // These are the virtual methods that the derived class
	      // must implement to be a Device object.
	  private:

	    virtual void ring_host(unsigned idx) =0;

	      // The base class assumes that mailbox 0 and 1 are input
	      // mailboxes, and 2 is an output mailbox. At any rate
	      // all must be distinct. How they are allocated to
	      // hardware depends on the derived class. The BUS class
	      // will only use the mailbox registers that way.

	    virtual void set_mailbox(unsigned idx, unsigned long value) =0;

	    virtual unsigned long get_mailbox(unsigned idx) const =0;

	      // A bus_transfer causes a DMA operation to be loaded
	      // into hardware and performed. The BUS object won't
	      // call this unless the bus_transfer_done() returns.

	    virtual void bus_transfer(unsigned long pci,
				      void*local,
				      unsigned long count,
				      bool wr_flag) =0;

	    virtual void bus_transfer_done() =0;

	  private:
				     
	    BUS*master_;

	  private: // Not implemented
	    Device(const Device&);
	    Device& operator= (const Device&);
      };

      void open_device(Device*dev)
      { dev_ = dev; dev_->device_init(this); }

    public:

	// Pass to the constructor an array of buffers to start
	// with. The application can use this to pass buffers, or may
	// pass 0 and use the install_buffers method to install them.
      explicit BUS();
      ~BUS();

	// This method writes data to a frame on the host. If the
	// frame doesn't exist, quietly ignore the copy. There are
	// only 16 possible frames, from 0 - 15.
      void copy_to_frame(unsigned fr, size_t off, const void*buf, size_t cnt)
	    { copy_frame_(fr, off, const_cast<void*>(buf), cnt, true); }

      void copy_from_frame(unsigned fr, size_t off, void*buf, size_t cnt)
	    { copy_frame_(fr, off, buf, cnt, false); }


    private:

      void copy_frame_(unsigned fr, size_t off,
		       void*buf, size_t cnt,
		       bool wr_flag);

	// The Device class uses these methods to communicate
	// spontaneous events from the physical bus to the channel
	// objects that are interested. Those spontaneous events are
	// the DMA values described next.

      friend class Device;
      void set_root_table_ptr_(unsigned long ptr);
      void set_status_message_(unsigned long msg);
      void set_channel_event_();
      void check_channel_status_(const Channel*);
      void consume_out_buffer_(Channel*);
      void consume_in_buffer_(Channel*);


	// These sync primitives wait on complex conditions that allow
	// the rather complex tasks to proceed. By waiting on complex
	// conditions instead of many simple conditions, I reduce
	// deadlock potential.

      ISync dispatch_sync_;

    private:
      friend class Channel;

	// Each channel can be opened and closed individually. An open
	// channel will be referenced in the following table.

      Channel * channels_;

	// There is exactly one device required to do all this stuff.
      Device * dev_;

	// This is the Primary PCI address of the root table.
      volatile unsigned long root_table_ptr_;

    private: // not implemented
      BUS(const BUS&);
      BUS&operator= (const BUS&);
};

/*
 * $Log: BUS.h,v $
 * Revision 1.29  1998/06/10 16:53:49  steve
 *  2way access to frames.
 *
 * Revision 1.28  1998/05/26 16:16:32  steve
 *  New channel protocol
 *
 */

#endif
