#ifndef __chips_I960RP_H
#define __chips_I960RP_H
/*
 * Copyright (c) 1996-1998 Picture Elements, Inc.
 *    Stephen Williams <steve@picturel.com>
 *
 *    This source code is free software; you can redistribute it
 *    and/or modify it in source code form under the terms of the GNU
 *    Library General Public License as published by the Free Software
 *    Foundation; either version 2 of the License, or (at your option)
 *    any later version. In order to redistribute the software in
 *    binary form, you will need a Picture Elements Binary Software
 *    License.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU Library General Public License for more details.
 *
 *    You should have received a copy of the GNU Library General Public
 *    License along with this program; if not, write to the Free
 *    Software Foundation, Inc.,
 *    59 Temple Place - Suite 330
 *    Boston, MA 02111-1307, USA
 *
 *    You should also have recieved a copy of the Picture Elements
 *    Binary Software License offer along with the source. This offer
 *    allows you to obtain the right to redistribute the software in
 *    binary (compiled) form. If you have not received it, contact
 *    Picture Elements, Inc.,
 *    777 Panoramic Way
 *    Berkeley, CA 94704.
 */
#ident "$Id: I960RP.h,v 1.6 1998/05/26 16:16:32 steve Exp $"

/*
 * This class is a driver for the PCI Interface of the I960RP
 * processor. The I960RP is not really a seperate chip, but a part of the
 * I960RP processor/bridge/PCI interface, but the functionality is self-
 * contained enough that I write it into a seperate device driver.
 *
 * The PCI Interface is a means for the embedded processor (me)
 * communicating with the host processor through the primary PCI bus.
 *
 * Built In SelfTest (BIST)
 *
 * This driver supports the BIST PCI register by allowing a thread to
 * do the BIST operations. The intent is that a thread enables the
 * BIST feature, then waits for the host to set the Start BIST
 * bit. The thread completes the BIST and sets a completion code,
 * which automatically turns of the Start BIST bit.
 */

# include  <chips/BUS.h>

/*
 * NOTE:
 * The I960RP needs to know the interrupt vectors for the MU
 * and the DMA controller. It uses the <ucr/param.h> header
 * file to get the correct definitions for these.
 */

class I960RP : public BUS::Device {

	// 
    public:
      explicit I960RP();
      ~I960RP();

	// BIST Support: bist_enable turns on/off the BIST Capable bit
	// in the BIST register. bist_finished(code) sets
	// the 4-bit completion code and turns off the BIST Start bit.
      void bist_enable(bool =true);
      void bist_finished(unsigned char);

	// This method waits for a BIST signal from the host. When one
	// arrives, one thread will be released to perform the test
	// and call bist_finished().
      void bist_wait() const;

	// These methods provide an interface to the DMA channels of
	// the i960RP. The trans method starts a transfer, and the
	// done method waits for a transfer to complete.

	// NOTE that the bus_transfer and bus_transfer_done methods
	// use channel 1. Locks synchronize access to the DMA
	// controllers so they may be shared.

      void dma_trans(unsigned ch, unsigned long pci, void*local,
		     unsigned long count, bool wr_flag =false);
      void dma_done(unsigned ch);

      unsigned long readl(unsigned long addr);
      void writel(unsigned long addr, unsigned long val);

	// These methods are the interface that the BUS class uses to
	// communicate with the host device driver.
    private:
      virtual void ring_host(unsigned idx);
      virtual unsigned long get_mailbox(unsigned) const;
      virtual void set_mailbox(unsigned, unsigned long);

      virtual void bus_transfer(unsigned long pci, void*local,
				unsigned long count, bool wr_flag);
    public:
      virtual void bus_transfer_done();

    private:

      ILock dma_sync_[3];
      mutable ILock bist_sync_;

      static void mu_isr_(I960RP*, unsigned);
      static void dma_isr_(I960RP*, unsigned);

    private: // Not implemented
      I960RP(const I960RP&);
      I960RP& operator= (const I960RP&);
};

#endif
/*
 * $Log: I960RP.h,v $
 * Revision 1.6  1998/05/26 16:16:32  steve
 *  New channel protocol
 *
 * Revision 1.5  1997/11/13 19:25:00  demo
 *  make wait for transfer public.
 *
 * Revision 1.4  1997/02/12 05:57:49  steve
 *  Support all three i960RP DMA channels.
 *
 * Revision 1.3  1996/10/28 03:45:12  steve
 *  License notice.
 *
 * Revision 1.2  1996/10/08 02:27:51  steve
 *  Add support for the BIST PCI feature.
 *
 * Revision 1.1  1996/10/04 23:34:24  steve
 *  Add a BUS driver for the ISE board.
 *
 */
