#ifndef __chips_PLX9060_H
#define __chips_PLX9060_H
/*
 * Copyright (c) 1996 Picture Elements, Inc.
 *    Stephen Williams
 *
 *    This source code is free software; you can redistribute it
 *    and/or modify it in source code form under the terms of the GNU
 *    Library General Public License as published by the Free Software
 *    Foundation; either version 2 of the License, or (at your option)
 *    any later version. In order to redistribute the software in
 *    binary form, you will need a Picture Elements Binary Software
 *    License.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU Library General Public License for more details.
 *
 *    You should have received a copy of the GNU Library General Public
 *    License along with this program; if not, write to the Free
 *    Software Foundation, Inc.,
 *    59 Temple Place - Suite 330
 *    Boston, MA 02111-1307, USA
 *
 *    You should also have recieved a copy of the Picture Elements
 *    Binary Software License offer along with the source. This offer
 *    allows you to obtain the right to redistribute the software in
 *    binary (compiled) form. If you have not received it, contact
 *    Picture Elements, Inc.,
 *    777 Panoramic Way
 *    Berkeley, CA 94704.
 */
#ident "$Id: PLX9060.h,v 1.7 1996/10/28 03:45:13 steve Exp $"
/*
 * $Log: PLX9060.h,v $
 * Revision 1.7  1996/10/28 03:45:13  steve
 *  License notice.
 *
 * Revision 1.6  1996/09/04 03:20:55  steve
 *  Change the BUS::Device class to not have any device specific
 *  stuff in it. Instead, the PLX9060 class is derived from it
 *  and implements virtual methods.
 *
 * Revision 1.5  1996/08/06 03:48:20  steve
 *  Move the DMA mode into defines, properly
 *  reset the DMA, and deal with an extern
 *  start/finish from the PLX chip.
 *
 * Revision 1.4  1996/07/25 22:25:23  steve
 *  Support DMA0 in either direction.
 *
 * Revision 1.3  1996/07/24 00:02:26  steve
 *  PLX9060 uses the ILock class for synchronization,
 *  do a more thorough job of initializing the PLX9060.
 *
 * Revision 1.2  1996/07/19 18:12:38  steve
 *  Add DMA handling.
 *
 * Revision 1.1  1996/07/15 20:32:02  steve
 *  Add the chips library, and the PLX9060 driver class.
 *
 */

# include <ucr/ILock.h>
# include <chips/BUS.h>

/*
 * This class is a driver for the PLX PCI9060 PCI interface chip,
 * local side. It is up to a derived class or application to implement
 * a policy or protocol for communicating with the host.
 */

class PLX9060  : public BUS::Device {

    public:
      explicit PLX9060(volatile void*base_addr);
      ~PLX9060();


	// Perform a DMA transfer from an address on the PCI bus to
	// the local memory. A thread calls this to start a
	// transfer. It may block if a DMA is already in progress. The
	// wr_flag tells the chip to WRITE TO THE PCI BUS.
    public:
      void dma0_transfer(unsigned long pci,
			 void*local, unsigned long count, bool wr_flag);
      void dma0_transfer_done();


    private:
	// The 8 mailbox registers can be read or written at any time,
	// although it is generally best to coordinate with the host
	// somehow.

      unsigned long get_mailbox(unsigned idx) const;
      void set_mailbox(unsigned idx, unsigned long value);

	// There are 32 different interrupt signals that can be sent
	// to the host. The idx passed here tells which interrupt to
	// set. In return, the host has 32 rings it can send here and
	// the ring_isr sets a function to handle it.

      void ring_host(unsigned idx);

    private:
      void bus_transfer(unsigned long pci,
			 void*local, unsigned long count, bool wr_flag);
      void bus_transfer_done();

    private:

      ILock dma0_done_flag_;
      ILock dma1_done_flag_;

      static void isr_(PLX9060*, unsigned);

    private:
      enum OFF_ {
	    PCI_LATENCY = 0x0d, // byte sized register
	    PCI_MIN_GNT = 0x3e, // byte sized register

	    MAILBOX0 = 0x0c0,
	    MAILBOX1 = 0x0c4,
	    MAILBOX2 = 0x0c8,
	    MAILBOX3 = 0x0cc,
	    MAILBOX4 = 0x0d0,
	    MAILBOX5 = 0x0d4,
	    MAILBOX6 = 0x0d8,
	    MAILBOX7 = 0x0dc,

	    BELLS_IN = 0x0e0,
	    BELLS_OUT= 0x0e4,

	    ICS      = 0x0e8,

	    DMA0_MODE       = 0x100,
	    DMA0_PCI_ADDR   = 0x104,
	    DMA0_LOCAL_ADDR = 0x108,
	    DMA0_TRANSFER   = 0x10c,
	    DMA0_DESCRIPTOR = 0x110,

	    DMA1_MODE       = 0x114,
	    DMA1_PCI_ADDR   = 0x118,
	    DMA1_LOCAL_ADDR = 0x11c,
	    DMA1_TRANSFER   = 0x120,
	    DMA1_DESCRIPTOR = 0x124,

	    DMA_CS = 0x128,
	    DMA_ARBIT0 = 0x12c,
	    DMA_ARBIT1 = 0x130,

	    REG_SIZE = 0x134
      };

      struct REGS_ {
	    unsigned char bytes[REG_SIZE];
      };

      volatile REGS_ *const regs_;

      void regs_set_ulong_(OFF_ where, unsigned long value)
      { *(unsigned long*)(regs_->bytes+where) = value; }

      unsigned long regs_get_ulong_(OFF_ where) const
      { return *(unsigned long*)(regs_->bytes+where); }

      void regs_set_uchar_(OFF_ where, unsigned char value)
      { *(unsigned char*)(regs_->bytes+where) = value; }

      unsigned long regs_get_uchar_(OFF_ where) const
      { return *(unsigned char*)(regs_->bytes+where); }
};


#endif
