#ifndef __chips_PLX9060p_H
#define __chips_PLX9060p_H
/*
 * Copyright (c) 1997 Picture Elements, Inc.
 *    Stephen Williams
 *
 *    This source code is free software; you can redistribute it
 *    and/or modify it in source code form under the terms of the GNU
 *    Library General Public License as published by the Free Software
 *    Foundation; either version 2 of the License, or (at your option)
 *    any later version. In order to redistribute the software in
 *    binary form, you will need a Picture Elements Binary Software
 *    License.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU Library General Public License for more details.
 *
 *    You should have received a copy of the GNU Library General Public
 *    License along with this program; if not, write to the Free
 *    Software Foundation, Inc.,
 *    59 Temple Place - Suite 330
 *    Boston, MA 02111-1307, USA
 *
 *    You should also have recieved a copy of the Picture Elements
 *    Binary Software License offer along with the source. This offer
 *    allows you to obtain the right to redistribute the software in
 *    binary (compiled) form. If you have not received it, contact
 *    Picture Elements, Inc.,
 *    777 Panoramic Way
 *    Berkeley, CA 94704.
 */
#ident "$Id: PLX9060p.h,v 1.9 1997/12/18 22:51:14 steve Exp $"
/*
 * $Log: PLX9060p.h,v $
 * Revision 1.9  1997/12/18 22:51:14  steve
 *  read bus descriptors
 *
 * Revision 1.8  1997/11/05 22:40:34  steve
 *  Add more DMA registers.
 *
 * Revision 1.7  1997/11/03 23:22:53  steve
 *  enum for DMA registers.
 *
 * Revision 1.6  1997/08/06 02:19:00  steve
 *  Constant references to constant memory.
 *
 * Revision 1.5  1997/08/02 04:28:59  steve
 *  Some standard constants.
 *
 * Revision 1.4  1997/04/14 05:39:57  demo
 *  Add support for DMA and words.
 *
 * Revision 1.3  1997/04/02 21:05:30  steve
 *  Various access methods.
 *
 * Revision 1.2  1997/03/28 05:45:42  steve
 *  Add methods to control IO pins.
 *
 * Revision 1.1  1997/03/27 00:08:53  steve
 *  Add the PLX9060 PCI side driver, and make
 *  the BIOS32 support short word config cycles.
 *
 */

/*
 * This class supports the PLX PCI9060 PCI side. (See the PLX9060
 * class for support of the local side.) Use this class to drive
 * devices that use the PCI9060 interface chip from the PCI side.
 */

# include  <stddef.h>
class BIOS32;

class PLX9060p {

    private:
      volatile unsigned long&dword_(volatile void*base, unsigned off)
      { return * (volatile unsigned long*)((char*)base + off); }

      const volatile unsigned long&dword_(const volatile void*base,
					  unsigned off) const
      { return * (const volatile unsigned long*) ((const char*)base + off); }

      volatile unsigned char&byte_(volatile void*base, unsigned off)
      { return * ((volatile unsigned char*)base + off); }

      volatile unsigned short&word_(volatile void*base, unsigned off)
      { return * ((volatile unsigned short*)base + off/2); }

      const volatile unsigned short&word_(const volatile void*base,
					  unsigned off) const
      { return * ((const volatile unsigned short*)base + off/2); }

    public:
      explicit PLX9060p (const BIOS32*pci, unsigned char bus,
			 unsigned char dev_fn);
      ~PLX9060p();

	// Write or read mailbox registers of the PLX. The PLX has 8
	// of these, numbered from 0 - 7.
      unsigned long mailbox(unsigned idx, unsigned long val)
      { return (dword_(plx_base_, 0x40+idx*4) = val); }

      unsigned long mailbox(unsigned idx) const
      { return dword_(plx_base_, 0x40+idx*4); }


	// There are two pins on the PLX chip that support random
	// things, including LEDS, etc. These methods write a value to
	// the single output pin, or read values from the input pin.
      bool user_out(bool bit)
      {
	    if (bit) dword_(plx_base_, 0x6c) |= 0x10000;
	    else dword_(plx_base_, 0x6c) &= ~0x10000;
	    return bit;
      }

      bool user_out() const
      { return 0 != (dword_(plx_base_, 0x6c) & 0x10000); }

      bool user_in() const
      { return 0 != (dword_(plx_base_, 0x6c) & 0x20000); }


	// Use reset_peripheral(true) to force the peripherals on the
	// local side of the PLX into a reset state, and hold them
	// there. The reset will remain until later removed by calling
	// reset_peripheral(false).

      void reset_peripheral(bool flag)
      {
	    if (flag) dword_(plx_base_, 0x6c) |= 0x40000000;
	    else dword_(plx_base_, 0x6c) &= ~0x40000000;
      }


	// The PLX chip does not intrinsically make the DMA registers
	// available to the PCI side. THE EROM hack uses the expansion
	// rom region to gain access to the local bus of the PLX chip,
	// and then to the DMA registers from there. The DMA registers
	// will then be placed (by setting the EROM base address)
	// midway in the local bus region.
      void invoke_dma_erom_hack(unsigned long remap =0xe0008000);

	// These methods provide access to the local bus
	// configuration. They can (and should) be set from the EEPROM
	// instead, if possible.
      void set_bus_descriptors(unsigned long val)
      { dword_(plx_base_, 0x18) = val; }

      unsigned long get_bus_descriptors() const
      { return dword_(plx_base_, 0x18); }

      void set_bus_remap(unsigned long base)
      {
	    base = (base & ~3) | (dword_(plx_base_, 0x04) & 3);
	    dword_(plx_base_, 0x04) = base;
      }

      void bus_enable(bool flag)
      {
	    if (flag) dword_(plx_base_, 0x04) |=1;
	    else dword_(plx_base_, 0x04) &= ~1;
      }

      void set_bus_irqstatus(unsigned long val)
      { dword_(plx_base_, 0x68) = val; }

	// These registers access the local-side bus of the PLX, and
	// thus the logic being interfaced to.

      volatile unsigned long&bus_dword(unsigned off)
      { return dword_(usr_base_, off); }

      volatile unsigned short&bus_word(unsigned off)
      { return word_(usr_base_, off); }

      volatile unsigned char&bus_byte(unsigned off)
      { return byte_(usr_base_, off); }

      const volatile unsigned short&bus_word(unsigned off) const
      { return word_(usr_base_, off); }

	// The DMA registers are one of the following and are
	// accessible as dwords. See the PCI 9060 datasheet for
	// meaning of these registers.
    public:
      enum DMA_REG { DMA0_MODE=0x100,  DMA0_ADDR=0x104,
		     DMA0_LADDR=0x108, DMA0_COUNT=0x10c,
		     DMA0_DESC=0x110,  DMA1_MODE=0x114,
		     DMA1_ADDR=0x118,  DMA1_LADDR=0x11c,
		     DMA1_COUNT=0x120, DMA1_DESC=0x124,
		     DMA_CMD_STAT=0x128, DMA_ARB0=0x12c,
		     DMA_ARB1=0x130 };

      volatile unsigned long&dma_dword(DMA_REG off)
      { return dword_(dma_base_, off); }


	// Access the NVRAM that PLX chips use to store configuration
	// stuff. It is up to the caller to know that the memory is
	// the right size. The data sizes are in units of 16bit words,
	// and the addr is a word offset.
    public:
      bool nvram_present() const
      { return dword_(plx_base_, 0x6c) & 0x10000000; }

      void nvram_read(unsigned addr, unsigned short*data, size_t ndata);
      void nvram_write(unsigned addr, const unsigned short*data, size_t ndata);

    private:
      const BIOS32*const pci_;
      const unsigned char bus_, dev_fn_;

      volatile void*plx_base_;
      volatile void*usr_base_;
      volatile void*dma_base_;

      void nvram_send_bits_(unsigned long bits, unsigned nbits);

    private: // Not implemented
      PLX9060p(const PLX9060p&);
      PLX9060p& operator= (const PLX9060p&);
};

#endif
