#ifndef __chips_PM2m_H
#define __chips_PM2m_H
/*
 * Copyright (c) 1998 Picture Elements, Inc.
 *    Stephen Williams (steve@picturel.com)
 *
 *    This source code is free software; you can redistribute it
 *    and/or modify it in source code form under the terms of the GNU
 *    General Public License as published by the Free Software
 *    Foundation; either version 2 of the License, or (at your option)
 *    any later version. In order to redistribute the software in
 *    binary form, you will need a Picture Elements Binary Software
 *    License.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *  ---
 *    You should also have recieved a copy of the Picture Elements
 *    Binary Software License offer along with the source. This offer
 *    allows you to obtain the right to redistribute the software in
 *    binary (compiled) form. If you have not received it, contact
 *    Picture Elements, Inc., 777 Panoramic Way, Berkeley, CA 94704.
 */
#ident "$Id: PM2m.h,v 1.4 1998/06/17 01:30:41 steve Exp $"

/*
 * This is a generic driver for the Pixel Magic PM-2m High Speed Image
 * Processor. It provides a buffer-to-buffer processing model that is
 * much more convenient then hunting around for the various registers.
 *
 * The implementations herin are based on information in the
 *
 *    PM-2m Family User's Guide and
 *    PM-2m Programming Guide
 *
 * Available from Pixel Magic Incorporated <http://www.pixelmagic.com>
 *
 * This class is typically not enough to actually operate a board with
 * the chip, as parts like the DMA interface are elsewere. However,
 * any board with this code can use this code.
 */

class PM2m {

      enum Reg { CTL_MASTER=0x01, CTL_DECODE=0x02, CTL_ENCODE=0x03,
		 ENC_CODE=0x04, CTL_ROT_SCALE=0x05, CTL_KREG_FILL=0x06,
		 CTL_INT_DMA=0x07, CTL_SPECIAL=0x08, X_SCALE=0x09,
		 Y_SCALE=0x0a, IP_CLIP_OFF=0x0b, IN_PIXELS=0x0c,
		 IP_CLIP_PAD=0x0d, IL_CLIP_OFF=0x0e, IN_LINES=0x0f,
		 PIXELS=0x11, LINES=0x12, OP_CLIP_OFF=0x13,
		 OP_CLIP_EXT=0x14, OL_CLIP_OFF=0x15, OL_CLIP_EXT=0x16,
		 OP_PAD_LM=0x17, OP_PAD_RM=0x18, OL_PAD_TM=0x19,
		 OL_PAD_BM=0x1a, PAD_CONTROL=0x1b, INIT_X_SCALE=0x1c,
		 INIT_Y_SCALE=0x1d, INPUT_CTL=0x1e, OUTPUT_CTL=0x1f,
		 STATUS1=0x32, STATUS2=0x33, CTL_JBIG=0x34,
		 STATUS3=0x36, CTL_ENH=0x3b, CTL_FIFO=0x3d,
		 OUT_DATA=0x42, EXT_RAM=0x8000 };

      enum {  // Bits from CTL_MASTER
	      //ENABLE   = 0x0001,
	    HF_DEC   = 0x0002,
	    HF_ENC   = 0x0004,
	    EN_SCR   = 0x0008,
	    COLOR_IN = 0x0100,
	    COLOR_OUT= 0x0200,
	    BYTE_IN  = 0x0400,
	    BYTE_OUT = 0x0800,
	    OUTPUT_BYTE_FLIP = 0x4000,

	      // Bits from CTL_INT_DMA
	    EN_DONE  = 0x0001,
	    EN_ERROR = 0x0002,
	    EN_RAW   = 0x0004,
	    I_DMA_EN = 0x0010,
	    O_DMA_EN = 0x0020,
	    I_STROBE_EN = 0x0040,
	    O_STROBE_EN = 0x0080,
	      // Bits from CTL_SPECIAL
	    I_FIFO_EN= 0x0001,
	    O_FIFO_EN= 0x0002,
	      // Bits from CTL_ENH
	    WORD_RAM = 0x0001,
	    FAST_RAM = 0x0002 };

      void write_(Reg r, unsigned short value)
      {	index_ = r;
        data_ = value;
      }

      unsigned short read_(Reg r) const
      { index_ = r;
        return data_;
      }

	// Read-modify-write is common enough that we can save the
	// overhead of setting the index_ register for the write. This
	// sets bits in SET and clears bit in CLR. This is also more
	// optimizable then otherwise.
      void set_clr_(Reg r, unsigned short set, unsigned short clr)
      { index_ = r;
        unsigned short tmp = data_;
	data_ = (tmp | set) & ~clr;
      }

    public:
	// Construct this object with references to the index register
	// and data register of the chip. This class uses these
	// references to manipulate the chip itself.

      PM2m(volatile unsigned short&index, volatile unsigned short&data);
      ~PM2m();

      unsigned short sram_data(unsigned idx) const
      { index_ = EXT_RAM + idx;
        return data_;
      }

      void sram_data(unsigned idx, unsigned short val)
      { index_ = EXT_RAM + idx;
        data_ = val;
      }

    public:
      void disable() { set_clr_(CTL_MASTER, 0, 1); }
      void enable() { set_clr_(CTL_MASTER, 1, 0); }

	// These methods set the geometry of the image coming in or
	// going out. The geometries are only needed where raw image
	// data are flowing. For example, for a compression operation
	// only the in geometry need be set.
      void geometry_in(unsigned short width, unsigned short height);
      void geometry_out(unsigned short width, unsigned short height);

	// Compress the input image using huffman encoding. The mode
	// specifies the desired encoding parameters.
      enum HE_MODE { TIFF2=0x0008, G3_1d=0x0501, G3_2d=0x507, G4=0x202 };
      void huff_encode(HE_MODE mode, unsigned short k=0);

	// This method sends data through the chip, clipping if the
	// geometry requests it. Both the in and out geometries are
	// required for this.
      void pass_thru();


    public:
	// These are accesses to various status registers.
      unsigned short status1() const { return read_(STATUS1); }
      unsigned short status2() const { return read_(STATUS2); }
      unsigned short status3() const { return read_(STATUS3); }

      unsigned short width()  const { return read_(IN_PIXELS); }
      unsigned short height() const { return read_(IN_LINES); }

      bool is_done() const    { return read_(STATUS1) & 0x800; }
      bool is_enabled() const { return read_(CTL_MASTER) & 1; }

    public:
	// These methods set or set various configuration options of
	// the chip. These methods must be called once before any of
	// the operations methods are invoked, as they describe the
	// hardware environment where the chip exists.

      enum SizeCode { SIZE_512=0, SIZE_1024=1, SIZE_2048=2,
		      SIZE_4096=3, SIZE_8192=4, SIZE_16384=5,
		      SIZE_32768=6 };

      enum BurstCode { BURST_NONE=0, BURST_1=1, BURST_2=7, BURST_4=2,
		       BURST_8=3, BURST_16=4, BURST_32=5, BURST_64=6 };

	// Configure the SRAM size, width and speed.

      void configure_sram(SizeCode size, bool fast =false, bool wide=true);

	// Configure the PM2m chip to do burst mode DMA. Turn off
	// external fifos, and turn on the LAST_BURST flag (unless out
	// is set to no burst size.)

      void set_burst_dma_mode(BurstCode in, BurstCode out);

	// Enable the /INT line going active on whichever of these
	// events. If true, that event will cause /INT to go active,
	// if false, not.

      void configure_int(bool done, bool error, bool raw);

    private:
      volatile unsigned short&index_;
      volatile unsigned short&data_;
};

/*
 * $Log: PM2m.h,v $
 * Revision 1.4  1998/06/17 01:30:41  steve
 *  More compression types.
 *
 * Revision 1.3  1998/06/16 23:50:40  steve
 *  Add image dimension query.
 *
 * Revision 1.2  1998/06/16 00:32:44  steve
 *  Proper reset in front of ccitt compression.
 *
 * Revision 1.1  1998/03/11 18:21:52  steve
 *  Add PM2m driver class.
 *
 */
#endif
