#ifndef __mc68322_GDMA_H
#define __mc68322_GDMA_H


/*
 * Copyright (c) 1995 Picture Elements Inc.
 *	Stephen Williams
 *
 *    This source code is free software; you can redistribute it
 *    and/or modify it in source code form under the terms of the GNU
 *    Library General Public License as published by the Free Software
 *    Foundation; either version 2 of the License, or (at your option)
 *    any later version. In order to redistribute the software in
 *    binary form, you will need a Picture Elements Binary Software
 *    License.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU Library General Public License for more details.
 *
 *    You should have received a copy of the GNU Library General Public
 *    License along with this program; if not, write to the Free
 *    Software Foundation, Inc.,
 *    59 Temple Place - Suite 330
 *    Boston, MA 02111-1307, USA
 *
 *    You should also have recieved a copy of the Picture Elements
 *    Binary Software License offer along with the source. This offer
 *    allows you to obtain the right to redistribute the software in
 *    binary (compiled) form. If you have not received it, contact
 *    Picture Elements, Inc.,
 *    777 Panoramic Way
 *    Berkeley, CA 94704.
 */
#ident "$Id: GDMA.h,v 1.5 1996/10/28 03:45:17 steve Exp $"
/*
 * $Log: GDMA.h,v $
 * Revision 1.5  1996/10/28 03:45:17  steve
 *  License notice.
 *
 * Revision 1.4  1995/11/08 10:08:00  steve
 *  Allow DMA transfers to be stopped.
 *
 * Revision 1.3  1995/10/17 18:30:19  steve
 *  Abandon the wait_for returning count,
 *  Add test_for() to know when transfers
 *  can be started.
 *
 * Revision 1.2  1995/10/17 16:58:51  steve
 *  Support large GDMA transfers by reloading the
 *  registers when terminal count is reached.
 *
 * Revision 1.1  1995/10/15 21:55:22  steve
 *  Add the GDMA class to manage the GDMA controller.
 *  This takes care of sync and operation.
 *
 */


/*
 * The GDMA device is actually a DMA controller that can be used by
 * external hardware to do transfer data to/from dynamic ram. This
 * class encapsulates the operation of the GDMA device, including
 * interrupt handling and synchronization. It also supports transfers
 * larger then the real GDMA by reloading the registers when
 * reasonable.
 *
 * INTERRUPTS:
 * This class doesn't actually register and ISR to deal with its
 * interrupts. The trouble is, handling DMA interrupts varies from
 * application, or even different devices within an application. To
 * deal with this, handing interrupts is left to the  programmer.
 *
 * The application, in an ISR, calls isr_check() to cause the GDMA
 * device to check the status of the DMA. This method will notice any
 * changes in state, or do any needed cleanup. The other methods can
 * then be used to restart the GDMA operation, or release the device,
 * or whatever.
 *
 * Thus, the canonical interrupt handler would call isr_check(), would
 * notice what is finished, and then would start the next transfer, if
 * necessary.
 *
 * NOTES:
 * Currently, only 16bit transfers are supported, and the transfer
 * count is given in 16bit words.
 *
 * Currently, only transfers from device to DRAM are supported, and
 * the to pointer passed to transfer must be in DRAM.
 */

# include  <ucr/ISync.h>
# include  <stdlib.h>

class GDMA {

        // Create the GDMA device, assign an interrupt level and
	// initialize.
    public:
      explicit GDMA(unsigned char level, unsigned long base = 0xfff000);

        // Set up and start a transfer from the device to DRAM. The
	// GMDA device must not be busy for this to work.
      void transfer(void*to, const void*from, size_t count);

        // Wait for the transfer to finish. If flush==true, then send
	// a flush signal to the GDMA before waiting.
      void wait_for(bool flush = false);
      bool test_for() const;

        // Use this method to stop the transfer early.
      void stop_transfer();

        // To get this class to notice changes in GDMA status, call
	// this method.
    public:
      void isr();

    private:
      ISync sync_;

      void setup_transfer_();
      size_t total_count_;
      unsigned short active_count_;
      char*active_to_;
      const void*active_from_;

      struct REGS_ {
	    void*volatile dram_address;
	    void*volatile sram_address;
	    volatile unsigned short count;
	    volatile unsigned short flush;
	    volatile unsigned short control;
      };

      struct INTR_ {
	    volatile unsigned short enable_flag;
	    volatile unsigned short int_flag;
	    volatile unsigned short level;
      };

      REGS_*const regs_;
      INTR_*const intr_;
      volatile unsigned short*const csdtr_;

    private: // Not implemented
      GDMA(const GDMA&);
      GDMA& operator= (const GDMA&);
};


#endif
