#ifndef __m68322_MSpeaker_H
#define __m68322_MSpeaker_H

/*
 * Copyright (c) 1995 Picture Elements Inc.
 *	Stephen Williams
 *
 *    This source code is free software; you can redistribute it
 *    and/or modify it in source code form under the terms of the GNU
 *    Library General Public License as published by the Free Software
 *    Foundation; either version 2 of the License, or (at your option)
 *    any later version. In order to redistribute the software in
 *    binary form, you will need a Picture Elements Binary Software
 *    License.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU Library General Public License for more details.
 *
 *    You should have received a copy of the GNU Library General Public
 *    License along with this program; if not, write to the Free
 *    Software Foundation, Inc.,
 *    59 Temple Place - Suite 330
 *    Boston, MA 02111-1307, USA
 *
 *    You should also have recieved a copy of the Picture Elements
 *    Binary Software License offer along with the source. This offer
 *    allows you to obtain the right to redistribute the software in
 *    binary (compiled) form. If you have not received it, contact
 *    Picture Elements, Inc.,
 *    777 Panoramic Way
 *    Berkeley, CA 94704.
 */
#ident "$Id: MSpeaker.h,v 1.2 1996/10/28 03:45:18 steve Exp $"
/*
 * $Log: MSpeaker.h,v $
 * Revision 1.2  1996/10/28 03:45:18  steve
 *  License notice.
 *
 * Revision 1.1  1995/12/12 00:01:48  steve
 *  Add speaker driver.
 *
 */

/*
 * Generate a square wave on the BG* line that can be used to drive a
 * speaker. The requested tone frequency is approximated and sent to
 * the speaker.
 *
 * HOW IT WORKS:
 * (Speaker support is not documented in the 322 manual, as not all
 * versions of the 322 support it. Your CPU must have a rom mask
 * higher then 2F65E.)
 *
 * The Speaker class uses the MC68322 buzzer driver. This driver has a
 * 16 bit down-counter and an enable bit at offset 0x608 in the 68322
 * registers.
 *
 * When enabled, the counter is loaded with the interval and counted
 * down by CLK1. When the counter reaches 0, the BG* line is toggled
 * and the counter reloaded. Thus, a square wave is generated on the
 * BG* line. The wavelength is 2*interval.
 *
 * The interval comes from the interval register. 8 bits from the
 * interval register are used as the high 8 bits of the 16 bit
 * interval counter, and 0xff as the low 8 bits. Thus, a value of 0x06
 * in the interval register gets an interval count of 0x6ff, and so a
 * wavelength of 2*0x6ff*Tclk1. With a 20MHZ CLK1, BG* is a square
 * wave with frequency 5.58KHZ.
 *
 * To enable this use of BG* (and disable its Bus Grant use), this
 * class sets bit 13 of the word at 0xfff910. This also disables the
 * default semantics ofAS* and DTACK*.
 *
 *
 * NOTE! This class only works for 68322 CPUs with mask revisions
 * higher then 2F65E. This device, although part of the 68322 (with
 * appropriate mask) is NOT documented in the MC68322 User's Manual.
 *
 * NOTE: If you use this class, then the AS*, DTACK* and BG* signals
 * will not be available for their documented purposes. The tone
 * generator uses the BG* signal, but the process of disabling that
 * line to use it, also disables the others. The constructor for this
 * class will set the alternate pin behavior.
 *
 * NOTE: Generating a tone does not consume CPU resources. It is done
 * in hardware without the attention of the 68EC000 (other then to set
 * up registers. So, generating a tone this way is very cheap.
 */

class MSpeaker {

    public:
      MSpeaker(unsigned long base =0xfff000,
	       unsigned long clk1frequency =20000000);

        // Generate a tone with the given frequency. The class will
	// derive the interval from the CLK1 frequency and the desired
	// HZ. If you ask for 0HZ, the speaker will be stopped.
      void tone(unsigned long hertz);


    private:
      const unsigned long clk1freq;

    private:
      struct REGS_ {
	    volatile unsigned short interval;
	    volatile unsigned short on_flag;
      };

      REGS_*const regs_;

    private:
      MSpeaker(const MSpeaker&);
      MSpeaker& operator= (const MSpeaker&);
};

#endif
