#ifndef __ucr_Selector_H
#define __ucr_Selector_H
/*
 * Copyright (c) 1996 Picture Elements, Inc.
 *    Stephen Williams
 *
 *    This source code is free software; you can redistribute it
 *    and/or modify it in source code form under the terms of the GNU
 *    Library General Public License as published by the Free Software
 *    Foundation; either version 2 of the License, or (at your option)
 *    any later version. In order to redistribute the software in
 *    binary form, you will need a Picture Elements Binary Software
 *    License.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU Library General Public License for more details.
 *
 *    You should have received a copy of the GNU Library General Public
 *    License along with this program; if not, write to the Free
 *    Software Foundation, Inc.,
 *    59 Temple Place - Suite 330
 *    Boston, MA 02111-1307, USA
 *
 *    You should also have recieved a copy of the Picture Elements
 *    Binary Software License offer along with the source. This offer
 *    allows you to obtain the right to redistribute the software in
 *    binary (compiled) form. If you have not received it, contact
 *    Picture Elements, Inc.,
 *    777 Panoramic Way
 *    Berkeley, CA 94704.
 */
#ident "$Id: Selector.h,v 1.3 1997/02/17 21:49:00 steve Exp $"
/*
 * $Log: Selector.h,v $
 * Revision 1.3  1997/02/17 21:49:00  steve
 *  reinterpret cast where 2.7.2 was too generous.
 *
 * Revision 1.2  1996/10/28 03:45:24  steve
 *  License notice.
 *
 * Revision 1.1  1996/08/30 23:38:34  steve
 *  Add the Selector class, and write into the Channel
 *  protocol some support for it.
 *
 */

/*
 * The technique these classes are designed to support allow an ISync
 * object to sleep on a bunch of conditions
 */

# include  <ucr/ISync.h>
# include  <assert.h>
class Selection;

/*
 * A selection is a condition that can be tested by Selectors, who
 * wait for the condition by expressing interest in the condition. The
 * Selector uses the test_it() method to test the state of the
 * condition. A thread that may change the state of the condition uses
 * the wakeup() method to notify all the interested Selectors of the
 * possible change.
 */

class Selection {

    public:
      struct CELL {
	    ~CELL() { next->prev = prev; prev->next = next; }
	    ISync*sync;
	    CELL *next;
	    CELL *prev;
      };

    public:
      Selection()
      {
	    list_.next = &list_;
	    list_.prev = &list_;
	    list_.sync = 0;
      }

      virtual ~Selection();
      virtual bool test_it() const =0;

      void wakeup()
      {
	    for (CELL*cur = list_.next ;  cur->sync ;  cur = cur->next)
		  cur->sync->wakeup();
      }

	// The Selector uses this method to announce to the Selection
	// object that an ISync is interested in the condition. The
	// Selection::wakeup() method will wake up all the ISync
	// objects in the cell list.

      void add_cell(CELL*cell) const
      {
	    assert(cell);
	    cell->next = &list_;
	    cell->prev = list_.prev;
	    cell->next->prev = cell;
	    cell->prev->next = cell;
      }

    private:
      mutable CELL list_;

    private: // Not implemented
      Selection(const Selection&);
      Selection& operator= (const Selection&);
};

/*
 * This is a slightly more convenient form of the Selection class, it
 * allows the programmer to write a function with a single argument
 * and use it as the condition to be tested.
 */
template <class ARG> class SleepFun  : public Selection {

    public:
      SleepFun(bool (*fun)(ARG), ARG arg)
      : fun_(fun), arg_(arg)
      { }

      bool test_it() const
      { return fun_(arg_); }

    private:
      bool (*const fun_)(ARG);
      const ARG arg_;
};


/*
 * The Selector is an ISync that can collect Selections and test them
 * en masse. A Selector can refer to SET_SIZE Selections. The program
 * set pointers to all the interesting selections, and calls
 * sleep(). The sleep() will block until at least one of the
 * Selections is true, and will remember all the true selections. The
 * program can then test all the results
 */

template <unsigned SET_SIZE> class Selector  : private ISync {

    public:
      Selector() { }

      void sleep()
      {
	    for (unsigned idx = 0 ;  idx < SET_SIZE ;  idx += 1)
		  mask_[idx] = false;

	    ISync::sleep(test_it_, this);
      }

      void set_selection(unsigned idx, const Selection*sel)
      {
	    assert(idx < SET_SIZE);
	    assert( sel );

	    table_[idx] = sel;
	    cells_[idx].sync = this;
	    mask_[idx] = false;

	    sel->add_cell(cells_+idx);
      }

      bool test_selection(unsigned idx) const
      {
	    assert( idx < SET_SIZE );
	    return mask_[idx];
      }

      void clear_selection(unsigned idx);

    private:
      static bool test_it_(Selector*obj)
      {
	    bool rc = false;
	    for (unsigned idx = 0 ;  idx < SET_SIZE ;  idx += 1)
		  if (obj->table_[idx] && obj->table_[idx]->test_it()) {
			obj->mask_[idx] = true;
			rc = true;
		  }

	    return rc;
      }

      const Selection*table_[SET_SIZE];
      Selection::CELL cells_[SET_SIZE];
      bool            mask_ [SET_SIZE];

    private: // Not implemented
      Selector(const Selector&);
      Selector& operator= (const Selector&);
};

#endif
