#ifndef __ucr_SyncRing_H
#define __ucr_SyncRing_H
/*
 * Copyright (c) 1997 Picture Elements, Inc.
 *    Stephen Williams (steve@picturel.com)
 *
 *    This source code is free software; you can redistribute it
 *    and/or modify it in source code form under the terms of the GNU
 *    General Public License as published by the Free Software
 *    Foundation; either version 2 of the License, or (at your option)
 *    any later version. In order to redistribute the software in
 *    binary form, you will need a Picture Elements Binary Software
 *    License.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *  ---
 *    You should also have recieved a copy of the Picture Elements
 *    Binary Software License offer along with the source. This offer
 *    allows you to obtain the right to redistribute the software in
 *    binary (compiled) form. If you have not received it, contact
 *    Picture Elements, Inc., 777 Panoramic Way, Berkeley, CA 94704.

 */
#ident "$Id: SyncRing.h,v 1.4 1998/07/09 22:17:27 steve Exp $"


# include  <ucr/RingBuffer.h>
# include  <ucr/ISync.h>
# include  <ucr/ucr.h>

/*
 * This class is like a Ring buffer, but it is synchronized. This
 * means that means that violations of content constraints leads to
 * blocking instead of confusion. This template is useful for
 * synchronizing cooperating threads.
 */
template <class T, unsigned size> class SyncRing {

    public:
      SyncRing() { }

	// Put the value into the queue. If there is no space, block
	// until there is. If there are any threads waiting on an
	// empty queue, wake them up.
      void put(T value)
      {
	    const spl_cookie_t cookie = thread_splhi();
	    sync_.sleep(wait_for_free_space_, this);
	    buf_.put(value);
	    thread_spl(cookie);
	    sync_.wakeup();
      }

	// Get a value from the queue. This always works, or blocks
	// until it will. If there are threads waiting for space to be
	// cleared in the ring, this will wake one up.
      T get()
      {
	    T result;
	    const spl_cookie_t cookie = thread_splhi();
	    sync_.sleep(wait_for_buffer_, this);
	    result = buf_.get();
	    thread_spl(cookie);
	    sync_.wakeup();
	    return result;
      }

      bool not_empty() const { return buf_.count() > 0; }
      bool not_full()  const { return buf_.space() > 0; }

      unsigned count() const { return buf_.count(); }

    private:
      ISync sync_;
      RingBuffer<T, size> buf_;

      static bool wait_for_free_space_(const volatile void*obj)
      { return reinterpret_cast<SyncRing*>(obj)->buf_.space() > 0; }

      static bool wait_for_buffer_(const volatile void*obj)
      { return reinterpret_cast<SyncRing*>(obj)->buf_.count() > 0; }

};

#endif
/*
 * $Log: SyncRing.h,v $
 * Revision 1.4  1998/07/09 22:17:27  steve
 *  Add a count method
 *
 * Revision 1.3  1998/06/26 03:58:47  steve
 *  Fix to handle stricter types.
 *
 * Revision 1.2  1998/04/21 16:21:40  steve
 *  Add not_empty test.
 *
 * Revision 1.1  1997/08/11 19:58:11  steve
 *  Add the SyncRing class.
 *
 */
