#ifndef __ucr_heap_H
#define __ucr_heap_H

/*
 * Copyright (c) 1995 Picture Elements Inc.
 *	Stephen Williams
 *
 *    This source code is free software; you can redistribute it
 *    and/or modify it in source code form under the terms of the GNU
 *    Library General Public License as published by the Free Software
 *    Foundation; either version 2 of the License, or (at your option)
 *    any later version. In order to redistribute the software in
 *    binary form, you will need a Picture Elements Binary Software
 *    License.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU Library General Public License for more details.
 *
 *    You should have received a copy of the GNU Library General Public
 *    License along with this program; if not, write to the Free
 *    Software Foundation, Inc.,
 *    59 Temple Place - Suite 330
 *    Boston, MA 02111-1307, USA
 *
 *    You should also have recieved a copy of the Picture Elements
 *    Binary Software License offer along with the source. This offer
 *    allows you to obtain the right to redistribute the software in
 *    binary (compiled) form. If you have not received it, contact
 *    Picture Elements, Inc.,
 *    777 Panoramic Way
 *    Berkeley, CA 94704.
 */
#ident "$Id: heap.h,v 1.7 1998/04/11 23:41:52 steve Exp $"
/*
 * $Log: heap.h,v $
 * Revision 1.7  1998/04/11 23:41:52  steve
 *  Add cell extent tests.
 *
 * Revision 1.6  1996/10/28 03:45:26  steve
 *  License notice.
 *
 * Revision 1.5  1995/10/17 21:27:00  steve
 *  Use stddef instead of stdlib,
 *  get stats for default heap.
 *
 * Revision 1.4  1995/10/15 23:59:18  steve
 *  Fix free cell splitting in the allocate,
 *  Add free space statistic,
 *  Make new() operators thread safe.
 *
 * Revision 1.3  1995/10/13 22:55:39  steve
 *  Add heap_stats function.
 *
 * Revision 1.2  1995/09/08 05:57:44  steve
 *  Better integration of heap handling with
 *  the new/delete operators.
 *
 * Revision 1.1  1995/09/02 18:27:12  steve
 *  Initial import of UCR core to cvs.
 *
 *
 */

# include  <stddef.h>

#if defined(__cplusplus)
# include  <ucr/ucr.h>
extern "C" {
#endif

/*
 * Manage hunks of memory as dynamic allocation heaps. Any chunk of
 * memory can be managed this way, although a default chunk will be
 * assigned by the C/C++ runtime for use by malloc/free.
 *
 * Memory allocated using heap_alloc contains enough context that the
 * heap_free function can figure out which HEAP_SPACE to return it
 * to. This special property allows one to use new(heap) to allocate
 * objects from specific heaps.
 */

struct HEAP_SPACE;

extern HEAP_SPACE* heap_init(void*space, unsigned long size);
extern void*heap_alloc(unsigned long size, HEAP_SPACE*space);
extern void heap_free(void*memory);

extern void heap_self_test(HEAP_SPACE*space);

/*
 * This function tests to see if the pointer is really within the
 * extent of the bit of memory allocated from the heap. If it is, and
 * there were no other problems, return 0. Otherwise, return a
 * non-zero value to indicate an error. If the return code is <0, the
 * error is a fault in the memory subsystem or the memory pointer.
 */
extern int heap_test_ext(const void*memory, const void*ptr);

#if defined(__cplusplus)
}
#endif


#if defined(__cplusplus)
  // This is a new operator that can be used to allocate memory from a
  // specified heap. This is compatible with the delete operator.
inline void* operator new (size_t size, HEAP_SPACE*space)
{
      const unsigned short cookie = thread_splhi();
      void*result = heap_alloc(size, space);
      thread_spl(cookie);
      return result;
}

inline void* operator new[] (size_t size, HEAP_SPACE*space)
{
      const unsigned short cookie = thread_splhi();
      void*result = heap_alloc(size, space);
      thread_spl(cookie);
      return result;
}


/*
 * The brk is the place in the heap (relative the beginning) where the
 * tail free cell lives. It amounts to an approximate amount of space
 * used in the heap, including fragmentation. New allocation won't
 * necessarily move the brk point up, nor will a deallocation
 * necessarily move the point down; but they probably will and this
 * gives a good estimate of how much heap space is hot.
 *
 * The freespace is the number of space available in free cells
 * underneath the break point. It is not the limit on allocation as
 * long as the brk point can move up.
 */
extern void heap_stats(const HEAP_SPACE*space, size_t&brk, size_t&freespace);

#endif

#endif
