/*
 * Copyright (c) 1996 Picture Elements, Inc.
 *    Stephen Williams
 *
 *    This source code is free software; you can redistribute it
 *    and/or modify it in source code form under the terms of the GNU
 *    Library General Public License as published by the Free Software
 *    Foundation; either version 2 of the License, or (at your option)
 *    any later version. In order to redistribute the software in
 *    binary form, you will need a Picture Elements Binary Software
 *    License.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU Library General Public License for more details.
 *
 *    You should have received a copy of the GNU Library General Public
 *    License along with this program; if not, write to the Free
 *    Software Foundation, Inc.,
 *    59 Temple Place - Suite 330
 *    Boston, MA 02111-1307, USA
 *
 *    You should also have recieved a copy of the Picture Elements
 *    Binary Software License offer along with the source. This offer
 *    allows you to obtain the right to redistribute the software in
 *    binary (compiled) form. If you have not received it, contact
 *    Picture Elements, Inc.,
 *    777 Panoramic Way
 *    Berkeley, CA 94704.
 */
#ident "$Id: PLX9060.cc,v 1.10 1996/10/28 03:45:41 steve Exp $"
/*
 * $Log: PLX9060.cc,v $
 * Revision 1.10  1996/10/28 03:45:41  steve
 *  License notice.
 *
 * Revision 1.9  1996/09/04 03:20:58  steve
 *  Change the BUS::Device class to not have any device specific
 *  stuff in it. Instead, the PLX9060 class is derived from it
 *  and implements virtual methods.
 *
 * Revision 1.8  1996/08/06 03:48:21  steve
 *  Move the DMA mode into defines, properly
 *  reset the DMA, and deal with an extern
 *  start/finish from the PLX chip.
 *
 * Revision 1.7  1996/08/05 18:23:53  steve
 *  Add DMA channel 1, and symbolic constants.
 *
 * Revision 1.6  1996/08/02 23:14:25  steve
 *  Do not enable LSERR.
 *
 * Revision 1.5  1996/07/30 04:21:24  steve
 *  Make sure bell is a transition.
 *
 * Revision 1.4  1996/07/25 22:25:25  steve
 *  Support DMA0 in either direction.
 *
 * Revision 1.3  1996/07/24 00:02:27  steve
 *  PLX9060 uses the ILock class for synchronization,
 *  do a more thorough job of initializing the PLX9060.
 *
 * Revision 1.2  1996/07/19 18:12:22  steve
 *  Add DMA handling.
 *
 * Revision 1.1  1996/07/15 20:32:05  steve
 *  Add the chips library, and the PLX9060 driver class.
 *
 */

# include  <chips/PLX9060.h>
# include  <ucr/ucr.h>
# include  <assert.h>

# define ICS_BELL_ACTIVE 0x00100000
# define ICS_DMA0_ACTIVE 0x00200000
# define ICS_DMA1_ACTIVE 0x00400000
# define ICS_BIST_ACTIVE 0x00800000

/*
 * These bits are masks into the DMA_CS register of the PLX
 * chip. There are also a few common mask combinations here.
 */
const unsigned long DMA_CS_0_ENABLE = 0x0001;
const unsigned long DMA_CS_0_START  = 0x0002;
const unsigned long DMA_CS_0_STOP   = 0x0004;
const unsigned long DMA_CS_0_CLRIRQ = 0x0008;
const unsigned long DMA_CS_0_DONE   = 0x0010;

const unsigned long DMA_CS_1_ENABLE = 0x0100;
const unsigned long DMA_CS_1_START  = 0x0200;
const unsigned long DMA_CS_1_STOP   = 0x0400;
const unsigned long DMA_CS_1_CLRIRQ = 0x0800;
const unsigned long DMA_CS_1_DONE   = 0x1000;

const unsigned long DMA_CS_x_STOP   = DMA_CS_1_STOP|DMA_CS_0_STOP;
const unsigned long DMA_CS_x_CLRIRQ = DMA_CS_0_CLRIRQ|DMA_CS_1_CLRIRQ;
const unsigned long DMA_CS_x_DONE   = DMA_CS_0_DONE|DMA_CS_1_DONE;

const unsigned long DMA_CS_COOKIE   = DMA_CS_0_ENABLE|DMA_CS_1_ENABLE;

/*
 * These are various DMA MODE register bits. OR the desired flags
 * together to get the desired DMA mode setting.
 */
const unsigned long DMODE_16BIT    = 0x00000001;
const unsigned long DMODE_32BIT    = 0x00000003;
const unsigned long DMODE_READY_EN = 0x00000040;
const unsigned long DMODE_BURST_EN = 0x00000100;
const unsigned long DMODE_IDONE_EN = 0x00000400;


PLX9060::PLX9060(volatile void*base_addr)
: regs_( (struct REGS_*)base_addr )
{
      regs_set_ulong_(BELLS_OUT, 0);
      regs_set_ulong_(BELLS_IN, 0xffffffff);

	// XINT0 is number 2
      set_isr(2, isr_, this);

	// Reset the DMA controller, wait for it to accept being
	// reset, and leave it disabled for now.
      regs_set_ulong_(DMA_CS, DMA_CS_x_CLRIRQ|DMA_CS_x_STOP);
	//regs_set_ulong_(DMA_CS, DMA_CS_x_CLRIRQ);
      while (DMA_CS_x_DONE != (regs_get_ulong_(DMA_CS) & DMA_CS_x_DONE))
	      /* Busy wait */ ;

	// Turn on the Done interrupt enable, so that DMA completions
	// cause an interrupt.
      regs_set_ulong_(DMA0_MODE, DMODE_32BIT
		                |DMODE_READY_EN
		                |DMODE_BURST_EN
		                |DMODE_IDONE_EN);
      regs_set_ulong_(DMA1_MODE, DMODE_32BIT
		                |DMODE_READY_EN
		                |DMODE_BURST_EN
		                |DMODE_IDONE_EN);

	// This enables the DMA channels. For some strange reason, the
	// PLX DMA controller never executes the first START after
	// being enabled, but still generates the interrupt. The start
	// just causes an immediate completion with no actual
	// action. The _STARTs here execute that mysterious NULL
	// transfer so that future transfers can ignore it.
      regs_set_ulong_(DMA_CS, DMA_CS_0_START|DMA_CS_1_START|DMA_CS_COOKIE);

	// Enable doorbell and DMA interupts from the host, and
	// doorbell interrupts to the host.
      regs_set_ulong_(ICS, 0x000f0300);

      regs_set_ulong_(DMA_ARBIT0, 0);
      regs_set_ulong_(DMA_ARBIT1, 0);

}

PLX9060::~PLX9060()
{
	// Disable interrupts from the PLX chip. This practically
	// turns it off.
      regs_set_ulong_(ICS, 0x00010100);
}

unsigned long PLX9060::get_mailbox(unsigned idx) const
{
      return *(unsigned long*)(regs_->bytes+MAILBOX0+idx*4);
}

void PLX9060::set_mailbox(unsigned idx, unsigned long value)
{
      *(unsigned long*)(regs_->bytes+MAILBOX0+idx*4) = value;
}

void PLX9060::ring_host(unsigned idx)
{
      unsigned long mask = 1 << idx;
      while (regs_get_ulong_(BELLS_OUT) & mask) /* busy wait */ ;
      regs_set_ulong_(BELLS_OUT, mask);
}

/*
 * Set up and start a DMA transfer using the specified channel. If the
 * channel is busy, wait until it is free for use, then start it on
 * mine.
 */
void PLX9060::dma0_transfer(unsigned long pci, void*local,
			    unsigned long count, bool wr_flag)
{
      unsigned long desc = 0x04;
      if (wr_flag) desc |= 0x08;

      dma0_done_flag_.get();

      regs_set_ulong_(DMA0_PCI_ADDR, pci);
      regs_set_ulong_(DMA0_LOCAL_ADDR, (unsigned long)local);
      regs_set_ulong_(DMA0_TRANSFER, count);
      regs_set_ulong_(DMA0_DESCRIPTOR, desc);
      regs_set_ulong_(DMA_CS, DMA_CS_0_START|DMA_CS_COOKIE);
}

void PLX9060::dma0_transfer_done()
{
      dma0_done_flag_.wait();
}

void PLX9060::bus_transfer(unsigned long pci, void*local,
			    unsigned long count, bool wr_flag)
{
      unsigned long desc = 0x04;
      if (wr_flag) desc |= 0x08;

      dma1_done_flag_.get();

      regs_set_ulong_(DMA1_PCI_ADDR, pci);
      regs_set_ulong_(DMA1_LOCAL_ADDR, (unsigned long)local);
      regs_set_ulong_(DMA1_TRANSFER, count);
      regs_set_ulong_(DMA1_DESCRIPTOR, desc);
      regs_set_ulong_(DMA_CS, DMA_CS_1_START|DMA_CS_COOKIE);
}

void PLX9060::bus_transfer_done()
{
      dma1_done_flag_.wait();
}


/*
 * Handle interrupts from the PLX chip. These include the doorbell
 * and DMA interrupts.
 */

void PLX9060::isr_(PLX9060*obj, unsigned)
{
      unsigned long flags = obj->regs_get_ulong_(ICS);

      if (ICS_DMA0_ACTIVE & flags) {
	    unsigned long cs = obj->regs_get_ulong_(DMA_CS);
	    obj->regs_set_ulong_(DMA_CS, DMA_CS_0_CLRIRQ|DMA_CS_COOKIE);

	    if (obj->dma0_done_flag_.got() && (cs & DMA_CS_0_DONE))
		  obj->dma0_done_flag_.put();
      }

      if (ICS_DMA1_ACTIVE & flags) {
	    unsigned long cs = obj->regs_get_ulong_(DMA_CS);
	    obj->regs_set_ulong_(DMA_CS, DMA_CS_1_CLRIRQ|DMA_CS_COOKIE);

	    if (obj->dma1_done_flag_.got() && (cs & DMA_CS_1_DONE))
		  obj->dma1_done_flag_.put();
      }

     
	// Test for all the possible doorbell interrupts. If there is
	// at least one, then dispatch it to the function set by the
	// application.

      if ( (flags = obj->regs_get_ulong_(BELLS_IN)) ) {
	    obj->regs_set_ulong_(BELLS_IN, flags);
	    obj->bus_bell_isr(flags);
      }
}
