/*
 * Copyright (c) 1998 Picture Elements, Inc.
 *    Stephen Williams (steve@picturel.com)
 *
 *    This source code is free software; you can redistribute it
 *    and/or modify it in source code form under the terms of the GNU
 *    General Public License as published by the Free Software
 *    Foundation; either version 2 of the License, or (at your option)
 *    any later version. In order to redistribute the software in
 *    binary form, you will need a Picture Elements Binary Software
 *    License.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *  ---
 *    You should also have recieved a copy of the Picture Elements
 *    Binary Software License offer along with the source. This offer
 *    allows you to obtain the right to redistribute the software in
 *    binary (compiled) form. If you have not received it, contact
 *    Picture Elements, Inc., 777 Panoramic Way, Berkeley, CA 94704.
 */
#ident "$Id: PLX9080p_nvram.cc,v 1.2 1998/04/25 05:07:16 steve Exp $"


# include  <chips/PLX9080p.h>
# include  <assert.h>

# define PLX_CTRL 0x6c

# define EEPROM_CLK 0x1000000
# define EEPROM_CS  0x2000000
# define EEPROM_TX  0x4000000
# define EEPROM_RX  0x8000000

inline void plx_writel(volatile void*base, unsigned off, unsigned long val)
{
      volatile unsigned long*tmp = (volatile unsigned long*)((char*)base + off);
      *tmp = val;
}


void PLX9080p::nvram_send_bits_(unsigned long bits, unsigned nbits)
{
      unsigned long mask = dword_(plx_base_, PLX_CTRL);
      mask &= ~(EEPROM_CLK|EEPROM_TX);
      mask |= EEPROM_CS;

      bits = bits << (32 - nbits);

      while (nbits > 0) {
	    if (bits & 0x80000000) {
		  plx_writel(plx_base_, PLX_CTRL, mask|EEPROM_TX);
		  plx_writel(plx_base_, PLX_CTRL, mask|EEPROM_TX|EEPROM_CLK);
	    } else {
		  plx_writel(plx_base_, PLX_CTRL, mask);
		  plx_writel(plx_base_, PLX_CTRL, mask|EEPROM_CLK);
	    }

	    bits <<= 1;
	    nbits -= 1;
	    plx_writel(plx_base_, PLX_CTRL, mask);
      }
}

void PLX9080p::nvram_read(unsigned addr, unsigned short*datap,
			  size_t ndata, unsigned naddr)
{
      unsigned long bits = dword_(plx_base_, PLX_CTRL);

      bits  &= ~(EEPROM_CLK|EEPROM_TX|EEPROM_CS);
      plx_writel(plx_base_, PLX_CTRL, bits);

      bits |= EEPROM_CS;
      plx_writel(plx_base_, PLX_CTRL, bits);

	// The read opcode is 110. Put this in front of the bits of
	// the address.
      unsigned long cmd = 6;
      cmd = (cmd << naddr) | addr;
      nvram_send_bits_(cmd, naddr+3);

	// Make sure the output is in a sane state
      plx_writel(plx_base_, PLX_CTRL, bits);

      while (ndata > 0) {
	    unsigned short data = 0;

	      // Shift 16 bits out of the EEPROM and into myself.
	    for (unsigned idx = 0 ;  idx < 16 ;  idx += 1) {
		  data <<= 1;

		  plx_writel(plx_base_, PLX_CTRL, bits|EEPROM_CLK);

		  if (dword_(plx_base_, PLX_CTRL) & EEPROM_RX) data |= 1;

		  plx_writel(plx_base_, PLX_CTRL, bits);
	    }


	    *(datap++) = data;
	    ndata -= 1;
      }

	// Remove the chip select.
      bits &= ~EEPROM_CS;
      plx_writel(plx_base_, PLX_CTRL, bits);
}


void PLX9080p::nvram_write(unsigned addr, const unsigned short*datap,
			   size_t ndata, unsigned naddr)
{
      unsigned long bits = dword_(plx_base_, PLX_CTRL);
      bits  &= ~ (EEPROM_CLK|EEPROM_CS|EEPROM_TX);
      plx_writel(plx_base_, PLX_CTRL, bits);

      bits |= EEPROM_CS;
      plx_writel(plx_base_, PLX_CTRL, bits);


	// write the Unlock command.
      nvram_send_bits_(0x13 << (naddr-2), naddr+3);

	// Pulse the CS to latch the command into the part.
      plx_writel(plx_base_, PLX_CTRL, bits&~EEPROM_CS);
      plx_writel(plx_base_, PLX_CTRL, bits);

      while (ndata > 0) {
	    plx_writel(plx_base_, PLX_CTRL, bits);

	      // Form the write command, with the address and the data.
	    unsigned long str = 0x5;
	    str <<= naddr;
	    str |= addr;
	    str <<= 16;
	    str |= (*datap & 0xffff);
	    nvram_send_bits_(str, 19+naddr);

	      // Lower the CS to start the load cycle.
	    plx_writel(plx_base_, PLX_CTRL, bits&~EEPROM_CS);

	    datap += 1;
	    ndata -= 1;
	    addr  += 1;

	      // This waits for the write to complete.
	    plx_writel(plx_base_, PLX_CTRL, bits);
	    while (0 == (dword_(plx_base_, PLX_CTRL) & EEPROM_RX))
		    /* busy wait */ ;

	      // Lowever the CS one more time to get out of the wait
	      // cycle.
	    plx_writel(plx_base_, PLX_CTRL, bits&~EEPROM_CS);
      }

	// Relock the chip with the bits 100 00XXXX(XX)
      plx_writel(plx_base_, PLX_CTRL, bits);
      nvram_send_bits_(4 << naddr, naddr+3);
      plx_writel(plx_base_, PLX_CTRL, bits&~EEPROM_CS);
}

/*
 * $Log: PLX9080p_nvram.cc,v $
 * Revision 1.2  1998/04/25 05:07:16  steve
 *  Support 93CS56 by adding address length parameter.
 *
 * Revision 1.1  1998/03/05 03:08:11  steve
 *  Add the PLX9080p class.
 *
 */

