/*
 * Copyright (c) 1996 Picture Elements, Inc.
 *    Stephen Williams
 *
 *    This source code is free software; you can redistribute it
 *    and/or modify it in source code form under the terms of the GNU
 *    Library General Public License as published by the Free Software
 *    Foundation; either version 2 of the License, or (at your option)
 *    any later version. In order to redistribute the software in
 *    binary form, you will need a Picture Elements Binary Software
 *    License.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU Library General Public License for more details.
 *
 *    You should have received a copy of the GNU Library General Public
 *    License along with this program; if not, write to the Free
 *    Software Foundation, Inc.,
 *    59 Temple Place - Suite 330
 *    Boston, MA 02111-1307, USA
 *
 *    You should also have recieved a copy of the Picture Elements
 *    Binary Software License offer along with the source. This offer
 *    allows you to obtain the right to redistribute the software in
 *    binary (compiled) form. If you have not received it, contact
 *    Picture Elements, Inc.,
 *    777 Panoramic Way
 *    Berkeley, CA 94704.
 */
#ident "$Id: GDB.cc,v 1.14 1998/06/30 00:00:21 steve Exp $"

# include  <gdb/GDB.h>
# include  <gdb/gdb_if.h>
# include  <ucr/ucr.h>
# include  <ucr/fault.h>
# include  <limits.h>
# include  <string.h>
# include  <stdlib.h>
# include  <assert.h>

static const unsigned char break_instruction[] = {0x00, 0x3e, 0x00, 0x66};
# define REGNO_IP 35

/*
 * This function flushes the instruction cache. Whenever I'm
 * reasonably sure text memory has been written to (for example when
 * setting breakpoints) this function should be called to be sure the
 * change actually takes affect for the processor.
 */
static inline void flush_instruction_cache()
{
      const unsigned long tmp = 0x100;
      asm volatile ("sysctl %0, 0, %0" : : "r" (tmp));
}

extern "C" void abort()
{
      for (;;) asm volatile ("cmpi 0,0; faulte");
}

GDB::GDB(BUS::Channel*port)
: self_(0), wait_wakeup_flag_(false), port_(port)
{
      port_->attention(reinterpret_cast<void(*)(void*,unsigned short,unsigned long)>(&attention_), this);
      for (unsigned idx = 0 ;  idx < FAULT_COUNT ;  idx += 1)
	    set_fault(idx, handle_fault_, this);
}

GDB::~GDB()
{
}

static bool test_tid_valid(THREAD*tid)
{
      THREAD*table[16];
      unsigned count = thread_list(table, 16);
      for (unsigned idx = 0 ;  idx < count ;  idx += 1)
	    if (table[idx] == tid) return true;

      return false;
}

bool GDB::wait_test_(GDB*obj)
{
      if (obj->wait_wakeup_flag_) return true;
      if (obj->trace_.peek()) return true;
      if (obj->faulty_.peek()) return true;
      return false;
}

void GDB::attention_(GDB*obj, unsigned short, unsigned long)
{
      obj->wait_wakeup_flag_ = true;
      obj->wait_sync_.wakeup();
}

void GDB::cmd_noop_(gdb_if&head, const void*)
{
      reply_(head, 0, 0);
}

void GDB::cmd_write_(gdb_if&head, const void*buffer)
{
      memcpy((void*)head.arg1, buffer, head.arg2);
      reply_(head, 0, 0);
}

/*
 * This command reads values from memory. If the value is an alogned
 * word, then read it with a wide access, otherwise accept a byte copy
 * to the transmit buffer. A word read is better handled by certain
 * memory-mapped devices.
 */
void GDB::cmd_read_(gdb_if&head, const void*)
{
      if ((0 == (head.arg1 % sizeof(unsigned long)))
	  && (head.arg2 == sizeof(unsigned long))) {

	    unsigned long val = *(unsigned long*)head.arg1;
	    reply_(head, &val, head.arg2);

      } else {
	    reply_(head, (void*)head.arg1, head.arg2);
      }
}

void GDB::cmd_threads_(gdb_if&head, const void*)
{
      THREAD*table[16];
      head.arg1 = thread_list(table, 16);
      reply_(head, table, head.arg1 * sizeof(THREAD*));

}

void GDB::cmd_get_regs_(gdb_if&head, const void*)
{
      char regs[36*4];
      THREAD*tid = (THREAD*)head.arg1;
      if (test_tid_valid(tid)) {
	    thread_regs(tid, regs);
	    reply_(head, regs, sizeof regs);
      } else {
	    reply_(head, 0, 0);
      }
}

void GDB::cmd_set_regs_(gdb_if&head, const void*)
{
      THREAD*tid = (THREAD*)head.arg1;
      if (test_tid_valid(tid)) thread_set_reg(tid, head.arg2, head.arg3);
      reply_(head, 0, 0);
}

void GDB::cmd_thread_test_(gdb_if&head, const void*)
{
      THREAD*tid = (THREAD*)head.arg1;
      head.arg1 = test_tid_valid(tid) ? 1 : 0;
      reply_(head, 0, 0);
}

/*
 * Wait for there to be at least one faulty thread, then suspend all
 * the threads and divide them into faulty and ready threads. Do not
 * pull faulty threads out of order from the faulty_ queue.
 */
void GDB::cmd_wait_(gdb_if&head, const void*)
{
      THREAD*table[16];

      if (head.arg1 == 0)
	    wait_sync_.sleep(reinterpret_cast<ISync::funp>(wait_test_), this);
      wait_wakeup_flag_ = false;

	// This goes through the list of available threads, and pulls
	// the ones that are not faulted into the thread_ list. Do
	// that safely. This suspends the entire application.

      spl_cookie_t cookie = thread_splhi();
      unsigned nthread = thread_list(table, 16);
      for (unsigned idx = 1 ;  idx < nthread ;  idx += 1) {
	    THREAD*tid = table[idx];
	    if (! get_fault(tid)) ready_.enqueue(tid);
      }
      thread_spl(cookie);


	// Now reply to the host debugger with information about the
	// first fault. If there are no faults (caused by a
	// non-blocking wait) then return the first ready thread.

      if (THREAD*fid = trace_.peek()) {
	    head.arg1 = (unsigned long)fid;
	    head.arg2 = FAULT_TRACE;

	      // Fix up all the threads that hit a breakpoint
	      // instruction by writing the correct ip value.
	    THREAD*tmp = fid;
	    do {
		  const FAULT_RECORD*frp = get_fault(tmp);
		  if (frp->subtype & 0x80)
			thread_set_reg(tmp, REGNO_IP, (unsigned long)frp->ip);
		  trace_.enqueue(tmp);
	    } while ( (tmp = trace_.peek()) != fid);

      } else if (THREAD*fid = faulty_.peek()) {
	    const FAULT_RECORD*frp = get_fault(fid);
	    head.arg1 = (unsigned long)fid;
	    head.arg2 = fault_type(frp);

      } else {
	    head.arg1 = (unsigned long)ready_.peek();
	    head.arg2 = ULONG_MAX;
      }

      reply_(head, 0, 0);
}


void GDB::cmd_resume_(gdb_if&head, const void*)
{
      spl_cookie_t cookie = thread_splhi();

      switch (head.arg1) {

	  default:
	      // If the debugger is asking to resume a specific
	      // thread, then put it is the ready queue. This will
	      // pull it out of the fault_ (or trace_) queue, which is a
	      // reasonable thing to do for some faults. The thread
	      // will not actually be run until later.
	      {
		    THREAD*tid = (THREAD*)head.arg1;
		    if (!test_tid_valid(tid)) break;
		    ready_.enqueue(tid, true);
	      }
	      break;

	  case -1:
	      // This is a special code for requesting all the
	      // threads, including those stopped for tracing, to be
	      // resumed. I manage that by making all the traced
	      // threads runnable and falling through.
	    for (THREAD*tid = trace_.pull() ; tid ; tid = trace_.pull())
		  thread_runnable(tid);

	  case 0:
	      // If I'm asked to start all the ready threads, go
	      // through the entire ready_ queue, pull the thread off
	      // the front and hand it back to the uCR scheduler.

	    for (THREAD*tid = ready_.pull() ; tid  ; tid = ready_.pull())
		  thread_runnable(tid);

	    break;
      }

      thread_spl(cookie);

      reply_(head, 0, 0);
}

/*
 * set_break_ writes the break_instruction into the program and
 * returns the memory it replaces. The host debugger remembers this
 * memory for me and sends it back when it clears the breakpoint.
 */
void GDB::cmd_set_break_(gdb_if&head, const void*)
{
      void*where = (void*)head.arg1;
      unsigned char buffer[sizeof break_instruction];
      memcpy(buffer, where, sizeof break_instruction);
      memcpy(where, break_instruction, sizeof break_instruction);

      flush_instruction_cache();

      reply_(head, buffer, sizeof break_instruction);
}

void GDB::cmd_clr_break_(gdb_if&head, const void*buffer)
{
      void*where = (void*)head.arg1;
      memcpy(where, buffer, sizeof break_instruction);

      flush_instruction_cache();

      reply_(head, 0, 0);
}

/*
 * If the input port has become ready to read, use this method to
 * receive the command and act on it.
 */
typedef void (GDB::*proc_method_t)(struct gdb_if&, const void*buffer);

void GDB::go()
{
      static const proc_method_t method_table[11] = {
	    &GDB::cmd_noop_,
	    &GDB::cmd_write_,
	    &GDB::cmd_read_,
	    &GDB::cmd_threads_,
	    &GDB::cmd_get_regs_,
	    &GDB::cmd_set_regs_,
	    &GDB::cmd_thread_test_,
	    &GDB::cmd_wait_,
	    &GDB::cmd_resume_,
	    &GDB::cmd_set_break_,
	    &GDB::cmd_clr_break_
      };

      self_ = thread_self();

      for (;;) {

	    gdb_if head;
	    char*buf = recv_(head);
	    head.ack = GIF_ACK_OK;

	    (this->*(method_table[head.cmd_code]))(head, buf);

	    if (buf) delete[]buf;
      }
}

bool GDB::handle_fault_(GDB*obj)
{
      bool continueable = false;
      const FAULT_RECORD*frp = get_fault(thread_self());
      if ((thread_self() == obj->self_) && (frp->ftype == FAULT_TRACE))
	    return true;

      if (obj->self_ == thread_self()) abort();

      spl_cookie_t cookie = thread_splhi();

      switch (frp->ftype) {
	  case FAULT_TRACE:
	    obj->trace_.enqueue(thread_self());
	    continueable = true;
	    break;
	  default:
	    obj->faulty_.enqueue(thread_self());
	    break;
      }

      obj->wait_sync_.wakeup();

      thread_reschedule();

      thread_spl(cookie);
      return continueable;
}


void GDB::pause()
{
      assert( self_ != thread_self() );

      const spl_cookie_t cookie = thread_splhi();

      trace_.enqueue(thread_self());
      wait_sync_.wakeup();
      thread_reschedule();

      thread_spl(cookie);
}

char*GDB::recv_(gdb_if&head)
{
      char*buf = 0;
      do {
	    port_->read(&head, sizeof head);
      } while ( (head.magic1 != GDB_IF_MAGIC1)
		&& (head.magic2 != GDB_IF_MAGIC2));

      if (head.ext) {
	    buf = new char [head.ext*4];
	    port_->read(buf, head.ext*4);
      }

      return buf;
}

void GDB::reply_(const gdb_if&src, const void*data, size_t ndata)
{
      gdb_if head = src;
      head.ext = (ndata + 3) / 4;
      port_->write(&head, sizeof head);
      if (head.ext) port_->write(data, head.ext*4);
}

/*
 * $Log: GDB.cc,v $
 * Revision 1.14  1998/06/30 00:00:21  steve
 *  Override abort instead of ucr-assert.
 *
 * Revision 1.13  1998/01/01 20:45:16  steve
 *  ISync has const methods.
 *
 * Revision 1.12  1997/08/26 19:24:47  steve
 *  trap asserts instead of aborts.
 *
 * Revision 1.11  1997/08/13 23:39:24  steve
 *  abort faults instead of halting.
 *
 * Revision 1.10  1997/08/13 22:33:22  steve
 *  Do word reads if it makes sense.
 *
 * Revision 1.9  1997/06/16 05:14:46  steve
 *  New channel protocol design and implementation.
 *
 * Revision 1.8  1997/02/17 21:49:01  steve
 *  reinterpret cast where 2.7.2 was too generous.
 *
 * Revision 1.7  1996/11/07 02:26:15  steve
 *  Use more portable methods.
 *
 * Revision 1.6  1996/10/31 00:40:44  steve
 *  Add pause for explicit breakpoints.
 *
 * Revision 1.5  1996/10/28 03:45:43  steve
 *  License notice.
 *
 * Revision 1.4  1996/09/09 23:52:23  steve
 *  Get GDB to notice attentions and wakeup wait.
 *
 * Revision 1.3  1996/09/03 21:35:40  steve
 *  Redesign the BUS class to use a packet protocol instead of
 *  the channel oriented protocol.
 *
 *  Remove use of the Selector from GDB, becuase the new
 *  BUS interface doesn't support it.
 *
 * Revision 1.2  1996/08/30 23:38:39  steve
 *  Add the Selector class, and write into the Channel
 *  protocol some support for it.
 *
 * Revision 1.1  1996/08/28 00:40:40  steve
 *  Add the GDB proxy to uCR.
 *
 */
