/*
 * Copyright (c) 1996 Picture Elements, Inc.
 *    Stephen Williams (steve@picturel.com)
 *
 *    This source code is free software; you can redistribute it
 *    and/or modify it in source code form under the terms of the GNU
 *    Library General Public License as published by the Free Software
 *    Foundation; either version 2 of the License, or (at your option)
 *    any later version. In order to redistribute the software in
 *    binary form, you will need a Picture Elements Binary Software
 *    License.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU Library General Public License for more details.
 *
 *    You should have received a copy of the GNU Library General
 *    Public License along with this program; if not, write to the
 *    Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *    Boston, MA 02111-1307, USA
 *
 *    You should also have recieved a copy of the Picture Elements
 *    Binary Software License offer along with the source. This offer
 *    allows you to obtain the right to redistribute the software in
 *    binary (compiled) form. If you have not received it, contact
 *    Picture Elements, Inc., 777 Panoramic Way, Berkeley, CA 94704.
 *
 */
#ident "$Id: vformat.cc,v 1.6 1997/08/17 01:17:01 steve Exp $"
/*
 * $Log: vformat.cc,v $
 * Revision 1.6  1997/08/17 01:17:01  steve
 *  Fix left padding of numbers.
 *
 * Revision 1.5  1996/10/25 23:16:20  steve
 *  Draw Inf and NaN values.
 *
 * Revision 1.4  1996/10/24 00:04:43  steve
 *  Add decimal formatting of floats.
 *
 * Revision 1.3  1996/09/17 19:24:28  steve
 *  More complete rendering of numbers.
 *
 * Revision 1.2  1996/09/16 19:34:32  steve
 *  Support sprintf.
 *
 * Revision 1.1  1996/09/16 17:02:34  steve
 *  Add fprintf and sprintf.
 *
 */

# include  "stdiox.h"
# include  <assert.h>
# include  <limits.h>
# include  <math.h>

/*
 * NOTE ABOUT MATH.H
 * I really really wanted to not include <math.h> (or any other
 * library) in the stdio implementation, but formatting of floating
 * point requires knowledge of floating point format and that would be
 * much worse then letting that live in a standard, well defined
 * library. Oh well.
 */

const unsigned PREC_UNDEFINED = UINT_MAX;

struct FormatFlags {
      unsigned base;

	// If the field width is > 0, the text is RIGHT justified.
	// If the field width is < 0, the text is LEFT justified.
      int field_width;
      unsigned precision;
      char sign_flag;
      char left_pad;
};

/*
 * The _vformat function is used both for fprintf and sprintf. The
 * higher level functions pass to _vformat a function pointer to be
 * used as fputc, the only output function used.
 */

static const char ldigits[] = "0123456789abcdef";
static const char udigits[] = "0123456789ABCDEF";

/*
 * This method draws an integer number to the output. Most of the
 * flags come from the FormatFlags structure, but a few are not knows
 * until the format letter and so are passed seperate.
 */
static unsigned draw_number(int (*putfun)(int c, FILE*s), FILE*str,
			    unsigned val, FormatFlags*flags, unsigned base,
			    bool base_ind)
{
      unsigned count = 0;
      char stack[32];
      unsigned stackp = 0;

      if (flags->precision == PREC_UNDEFINED) flags->precision = 1;

      while (val) {
	    stack[stackp++] = ldigits[val%base];
	    val /= base;
      }

	// Extend the number with leading zeros until the precision
	// requirement is met, ...
      while (stackp < flags->precision) stack[stackp++] = '0';

	// place any base indicator prefix onto the front of the string...
      if (base_ind) switch (base) {
	  case 16:
	    stack[stackp++] = 'x';
	    stack[stackp++] = '0';
	    break;

	  case 8:
	    if (stack[stackp-1] != '0') stack[stackp++] = '0';
	    break;
      }

	// Add any sign character necessary.
      if (flags->sign_flag) stack[stackp++] = flags->sign_flag;

	// pad on the left with spaces, as needed to fill the
	// field. If the number is already large enough to fill the
	// field, no pad will be written. If the field is to be right
	// justified, field_width will be <0, which is certainly less
	// then stackp and the loop will not be executed.

      while (flags->field_width > (int)stackp) {
	    putfun(flags->left_pad, str);
	    flags->field_width -= 1;
	    count += 1;
      }

	// Write the digits of the number, in correct order.
      count += stackp;
      while (stackp) putfun(stack[--stackp], str);

	// Pad on the right with spaces if needed to finish the
	// justification of the string. If the field is to be right
	// justified, it will be >0, so this loop will never be
	// executed.

      flags->field_width += count;
      while (flags->field_width < 0) {
	    putfun(' ', str);
	    flags->field_width += 1;
	    count += 1;
      }

      return count;
}

static unsigned draw_nan(int (*putfun)(int c, FILE*s), FILE*str,
			 double, FormatFlags*flags)
{
      unsigned count = 3;
      int field_width = flags->field_width;

      while (field_width > 3) {
	    putfun(' ', str);
	    field_width -= 1;
	    count += 1;
      }

      putfun('N', str);
      putfun('a', str);
      putfun('N', str);

      while (field_width < -3) {
	    putfun(' ', str);
	    field_width += 1;
	    count += 1;
      }

      return count;
}

static unsigned draw_inf(int (*putfun)(int c, FILE*s), FILE*str,
			 double val, FormatFlags*flags)
{
      unsigned count = 4;
      int field_width = flags->field_width;

      while (field_width > 4) {
	    putfun(' ', str);
	    field_width -= 1;
	    count += 1;
      }

      putfun((val < 0.0) ? '-' : '+', str);
      putfun('I', str);
      putfun('n', str);
      putfun('f', str);

      while (field_width < -4) {
	    putfun(' ', str);
	    field_width += 1;
	    count += 1;
      }

      return count;
}

/*
 * This function draws the floating point number in the form
 * [-]ddd.dddd however large the number is.
 */

static unsigned draw_decimal(int (*putfun)(int c, FILE*s), FILE*str,
			     double val, FormatFlags*flags)
{
      if (isnan(val)) return draw_nan(putfun, str, val, flags);
      if (!finite(val)) return draw_inf(putfun, str, val, flags);

      unsigned prec = flags->precision;
      if (prec == PREC_UNDEFINED) prec = 6;

      bool negative_flag = false;
      if (val < 0.0) {
	    negative_flag = true;
	    val = 0.0-val;
      }

      double integral = floor(val);

      char stack[64];
      unsigned stackp = 0;

	// Write the fraction part into the known length of the
	// precision. This only applies if the precision is greater
	// then zero. If the precision is zero, then leave off even
	// the decimal point.

      if (prec > 0) {
	    double fraction = val - integral;
	    unsigned fracp = prec + 1;
	    stackp = fracp;

	    stack[--fracp] = '.';
	    while (fracp > 0) {
		  fraction *= 10.0;
		  unsigned digit = unsigned(fraction);
		  stack[--fracp] = ldigits[digit];
		  fraction -= floor(fraction);
	    }
      }

	// Now write the integer part by pushing digits onto a
	// stack. Keep pushing until there is no more integral part.
      while (integral > 0) {
	    double next_integral = floor(integral/10.0);
	    stack[stackp++] = ldigits[unsigned(integral - next_integral*10)];
	    assert( stackp < sizeof stack );

	    integral = next_integral;
      }

	// Put the sign onto the number.
      if (negative_flag) stack[stackp++] = '-';


	// Finally, I have the number all made up. write it out by
	// reading it off the stack.

      unsigned count = stackp;
      while (stackp > 0) putfun(stack[--stackp], str);

      return count;
}

int _vformat(int (*putfun)(int c, FILE*s), FILE*str,
	     const char*format, va_list arg)
{
      unsigned count = 0;

      for (;;) {

	    while (*format && *format != '%') {
		  putfun(*format, str);
		  count += 1;
		  format += 1;
	    }

	      // This is the only exit from this loop. Here, I detect
	      // that I have come to the end of the format string.
	    if (*format == 0) return count;

	      // At this point, I know that format is pointing to the
	      // '%' of a format operation, so I start parsing the
	      // format, which is most of what this function does.
	    format += 1;

	      // Get the flags part of the format. The flags can
	      // appear in any order. When I find a character that is
	      // not a flag, I am done.
	    FormatFlags flags;
	    flags.sign_flag = 0;
	    flags.left_pad = ' ';
	    bool justify_flag = false;
	    bool base_ind = false;
	    for (;;) {
		  switch (*format) {
		      case '+':
			flags.sign_flag = '+';
			break;
		      case '-':
			justify_flag = true;
			break;
		      case ' ':
			if (flags.sign_flag == 0) flags.sign_flag = ' ';
			break;
		      case '#':
			base_ind = true;
			break;
		      case '0':
			flags.left_pad = '0';
			break;
		      default:
			goto done_flags;
		  }
		  format += 1;
	    }
      done_flags:

	      // Field width comes after the flags, and is a decimal
	      // number. The initial, default value is 0. After
	      // reading the number, use the justify_flag to turn it
	      // negative if left justified.
	    flags.field_width = 0;
	    for (;;) {
		  switch (*format) {
		      case '0' ... '9':
			flags.field_width = 10*flags.field_width + (*format-'0');
			break;
		      case '*':
			flags.field_width = va_arg(arg, int);
			if (flags.field_width < 0) justify_flag = false;
			break;
		      default:
			goto done_width;
		  }
		  format += 1;
	    }
      done_width:

	    if (justify_flag) flags.field_width = 0-flags.field_width;

	      // Precision is seperated from the field width with a
	      // period, and is a decimal number.
	    flags.precision = PREC_UNDEFINED;
	    if (*format == '.') for (format += 1, flags.precision = 0;;) {
		  switch (*format) {
		      case '0' ... '9':
			flags.precision = 10*flags.precision + (*format-'0');
			break;
		      case '*': {
			    int tmp = va_arg(arg, int);
			    flags.precision = (tmp >= 0)? tmp : (0-tmp);
			    break;
		      }
		      default:
			goto done_prec;
		  }
		  format += 1;
	    }

      done_prec:

	      // an integer value may have been a short or long, and
	      // promoted to be passed as a paramter. C does that, but
	      // we theoretically lose some information in the
	      // promotion process.
	      // XXXX For now, I pay no attention to this flag

	    switch (*format) {
		case 'h':
		  format += 1;
		  break;
		case 'l':
		  format += 1;
		  break;
		case 'L':
		  format += 1;
		  break;
	    }

	      // All the previous stuff was just parameters for the
	      // real formatting operation, which is selected by a
	      // single letter, as below.

	    switch (*format) {
		case 'd':
		case 'i': {
		      int val = va_arg(arg, int);
		      if (val < 0) {
			    flags.sign_flag = '-';
			    val = 0-val;
		      }
		      count += draw_number(putfun, str, val, &flags, 10,
					   false);
		      break;
		}

		case 'o': {
		      unsigned val = va_arg(arg, unsigned);
		      count += draw_number(putfun, str, val, &flags,
					   8, base_ind);
		      break;
		}

		case 'u': {
		      unsigned long val = va_arg(arg, unsigned);
		      count += draw_number(putfun, str, val, &flags,
					   10, false);
		      break;
		}

		case 'p':
		case 'x':
		case 'X': {
		      unsigned long val = va_arg(arg, unsigned);
		      count += draw_number(putfun, str, val, &flags,
					   16, base_ind);
		      break;
		}

		case 'f': {
		      double val = va_arg(arg, double);
		      count += draw_decimal(putfun, str, val, &flags);
		      break;
		}

		case 'e':
		case 'E':
		case 'g':
		case 'G': {
		      double val = va_arg(arg, double);
		      count += draw_decimal(putfun, str, val, &flags);
		      break;
		}

		case 'c': {
		      int val = va_arg(arg, int);
		      putfun(val, str);
		      break;
		}

		case 's': {
		      char*val = va_arg(arg, char*);
		      char*cp;
		      for (cp = val; *cp ; cp += 1) putfun(*cp, str);
		      count += cp-val;
		      break;
		}

		case 'n': {
		      int*val = va_arg(arg, int*);
		      *val = count;
		      break;
		}

		case '%':
		  putfun('%', str);
		  break;
		default:
		  break;
	    }

	    if (*format) format += 1;
      }
}
