
/*
 * Copyright (c) 1995 Picture Elements Inc.
 *	Stephen Williams
 *
 *    This source code is free software; you can redistribute it
 *    and/or modify it in source code form under the terms of the GNU
 *    Library General Public License as published by the Free Software
 *    Foundation; either version 2 of the License, or (at your option)
 *    any later version. In order to redistribute the software in
 *    binary form, you will need a Picture Elements Binary Software
 *    License.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU Library General Public License for more details.
 *
 *    You should have received a copy of the GNU Library General Public
 *    License along with this program; if not, write to the Free
 *    Software Foundation, Inc.,
 *    59 Temple Place - Suite 330
 *    Boston, MA 02111-1307, USA
 *
 *    You should also have recieved a copy of the Picture Elements
 *    Binary Software License offer along with the source. This offer
 *    allows you to obtain the right to redistribute the software in
 *    binary (compiled) form. If you have not received it, contact
 *    Picture Elements, Inc.,
 *    777 Panoramic Way
 *    Berkeley, CA 94704.
 */
#ident "$Id: Mutex.cc,v 1.7 1996/10/28 03:46:10 steve Exp $"
/*
 * $Log: Mutex.cc,v $
 * Revision 1.7  1996/10/28 03:46:10  steve
 *  License notice.
 *
 * Revision 1.6  1996/08/24 20:57:31  steve
 *  The priority cookie is not necessarily short, so make
 *  a typedef for it.
 *
 *  There is no need (indeed no way) to save and restore the
 *  fp of a thread, so put the PFP, which is really the fp,
 *  into the fp register slot.
 *
 *  Slightly better check of the stack frame at thread switch
 *  time. Also have faults check for some bad things.
 *
 * Revision 1.5  1996/08/08 05:22:41  steve
 *  Get rid of the __Scheduler class, it was not
 *  needed. Add the thread_runnable and _reschedule
 *  functions for thread scheduling.
 *
 * Revision 1.4  1996/08/03 03:30:38  steve
 *  Create the ucr/asm.h header as a target specific
 *  header with line functions, etc; Replace the
 *  __isr_run_flag with a bit of inline assembly code.
 *
 *  Move much thread code out of the __Secheduler class
 *  and into more public thread_ functions, to eliminate
 *  some wasted function call depth. Even the main_thread
 *  is move into its own file.
 *
 *  Keep a list of available threads for debug use, and
 *  add a function to get thread registers.
 *
 * Revision 1.3  1995/10/20 04:00:43  steve
 *  No enter in ISRs.
 *
 * Revision 1.2  1995/10/09 06:56:22  steve
 *  Add missing methods.
 *
 * Revision 1.1  1995/09/02 18:27:22  steve
 *  Initial import of UCR core to cvs.
 *
 *
 */

# include  <ucr/Mutex.h>
# include  <ucr/ucr.h>
# include  <assert.h>

Mutex::Mutex()
: busy_flag(0)
{ }

Mutex::~Mutex()
{
      assert( !busy_flag );
}

void Mutex::enter()
{
      assert( ! isr_hot_flag() );
      spl_cookie_t cookie = thread_splhi();
      while (busy_flag) {
	    stuck_.enqueue(thread_self());
	    thread_reschedule();
      }

      busy_flag = 1;

      thread_spl(cookie);
}


void Mutex::leave()
{
      spl_cookie_t cookie = thread_splhi();

      busy_flag = 0;
      THREAD*next = stuck_.peek();
      if (next) thread_switch(next);
      else thread_reschedule();
      thread_spl(cookie);
}


Condition::Condition(Mutex&m)
: master_(m)
{
}

Condition::~Condition()
{
}

/*
 * Condition variable manipulations are protected by the Mutex, so I
 * need not protect them from ISRs.
 */

/*
 * Wait for another thread to signal me. I do this by putting myself
 * in the waiting_ queue and leaving(). The leave() will perform
 * rescheduling as needed, giving preference to threads trying to
 * enter the mutex.
 */
void Condition::wait()
{
      waiting_.enqueue(thread_self());
      master_.leave();
      master_.enter();
}

/*
 * If there is a thread waiting, change it from waiting_ to
 * stuck_. Then, when I leave it will be scheduled. Promptly reenter
 * the mutex before continuing myself.
 */
void Condition::signal()
{
      THREAD*next = waiting_.peek();
      if (next) {
	    master_.stuck_.enqueue(next, true);
	    master_.leave();
	    master_.enter();
      }
}


void Condition::broadcast()
{
      THREAD*next;
      while ( (next = waiting_.pull()) )
	master_.stuck_.enqueue(next, true);

      master_.leave();
      master_.enter();
}
