#!wishpci

# --
# Copyright (c) 1996 Picture Elements, Inc.
# Stephen Williams (steve@picturel.com)
#
# $Id: ise_board.tcl,v 1.6 1997/09/11 21:59:14 steve Exp $
#
# This command uses the PCICONF package (available from Picture Elements)
# to implement a control panel for the ISE board. It is intended to work
# with the bootprom program included with the ISE configuration of uCR.
#
# The RESET button resets the i960 on the ISE board, and causes the
# BOOTPROM to take over the CPU. It does not yet assure that there are
# no processes or device drivers holding the device, that is up to you
# for now.
#
#    This source code is free software; you can redistribute it
#    and/or modify it in source code form under the terms of the GNU
#    General Public License as published by the Free Software
#    Foundation; either version 2 of the License, or (at your option)
#    any later version. In order to redistribute the software in
#    binary form, you will need a Picture Elements Binary Software
#    License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
#
#    You should also have recieved a copy of the Picture Elements
#    Binary Software License offer along with the source. This offer
#    allows you to obtain the right to redistribute the software in
#    binary (compiled) form. If you have not received it, contact
#    Picture Elements, Inc., 777 Panoramic Way, Berkeley, CA 94704.
#
# $Log: ise_board.tcl,v $
# Revision 1.6  1997/09/11 21:59:14  steve
#  Run under windows nt.
#
# Revision 1.5  1997/06/15 20:18:14  steve
#  Use shared objects.
#
# Revision 1.4  1997/04/14 05:37:54  demo
#  Wrong device id. whoops.
#
# Revision 1.3  1997/03/25 00:23:21  steve
#  Parameterize PCI ids.
#
# Revision 1.2  1996/10/30 23:16:45  steve
#  Add status and sanity tests.
#
# Revision 1.1  1996/10/23 20:13:30  steve
#  Add a GUI control panel
#
#

# --
# The ATU and bridge are different function units of the same board.

# Load the pciconf extension. Try the prelinked version first, in case
# it is available. Otherwise, load the file explicitly.

if {"$tcl_platform(platform)" == "windows"} {
    set pcidev "\\\\.\\pciconf"
    load pciconf
} else {
    set pcidev "/dev/pciconf"
    if {[catch {load {} pciconf}]} {
	load /usr/local/lib/libpciconf.so
    }
}

set vendor_id 0x8086
set device_id 0x1960

pciopen ATU -r -w $pcidev
ATU select $vendor_id $device_id 0

pciopen bridge -r -w $pcidev
bridge select 0x8086 0x0960 0

# --
# See if the ISE module is busy. Use the ISE module reference count to
# determine this.

proc test_ise_busy {} {
    global tcl_platform

    # Note that windows is broken in this score. Hope for the best.
    if {"$tcl_platform(platform)" == "windows"} {return 0}

    # If the module is not even loaded, it is certainly not busy.
    if { 1 == [catch {set open_count [exec grep ise "/proc/modules"]}] } {
	return 0
    }

    # Get the reference count. If that comes to 0 then the device is
    # not busy so return 0.
    if { 0 == [lindex "$open_count" 2] } { return 0 }

    return 1
}

# --
# This procedure uses the bridge unit to perform a hard reset of the
# i960 core and its local bus. This causes it to restart the bootprom
# image, and allow me to download programs into it.

proc reset_to_bootprom {} {
    global vendor_id device_id pcidev

    if { 1 == [test_ise_busy] } {
	popup {Cannot reset the ISE board, because it is busy.\
	       Check that no processes have the board opened \
	       and try again.}
       return
   }

    # --
    # If the i960RP is not yet being held reset, then I must force it
    # reset myself. This will trash the configuration space registers for
    # the ATU, so I must remember to restore them.

    if {([bridge get_byte 0x40] & 0x02) == 0} {

	pciopen ATU -r -w $pcidev
	ATU select $vendor_id $device_id 0

	set reg_list "0x04 0x0c 0x10 0x30 0x3c"

	foreach idx $reg_list { set save($idx) [ATU get_dword $idx] }

	# This reset the local bus on the i960RP...
	bridge set_byte 0x40 0x20

	# and this restores the registers that the BIOS set up.
	foreach idx $reg_list { ATU set_dword $idx $save($idx) }
    }

    # --
    # All set. Release the CPU.
    bridge set_byte 0x40 0
}

# --
# This procedure uses the BIST flag to release the BOOTPROM to starting
# the application that has been loaded. If the BIST Capable bit is not
# set, then the program has not been successfully loaded and the start
# cannot proceed.

proc release_program {} {

    if { 1 == [test_ise_busy] } {
	popup {Cannot start the application on the ISE board,\
	       because it is busy. Check that no processes have\
	       the board opened, and try again.}
       return
   }

    set flags [ATU get_word 0xe]

    if { ! ($flags & 0x8000) } {
	popup {The application is not loaded (BIST Capable bit not set.)\
	       Load a linked program into the ISE device and try again.}
	return
    }

    ATU set_word 0xe [expr {$flags | 0x4000}]
}


label .title -text "ISE Board Control Panel"

button .reset -text "Reset to Bootprom" -command reset_to_bootprom
button .start -text "Start Application" -command release_program

pack .title .reset .start

# --
# Build up a dialog box for displaying interesting messages. The only
# interactions applicable are simple acknowledgements of messages.

set message_box ""

toplevel .db

frame   .db.frame
label   .db.frame.title \
	-width 32 \
	-text "ISE Control Panel Message"
message .db.frame.message \
	-justify left \
	-relief groove \
	-aspect 500 \
	-bd 2 \
	-textvariable message_box
button  .db.frame.dismiss \
	-text "Dismiss" \
	-command {wm withdraw .db ; grab release .db}

pack .db.frame.title
pack .db.frame.message -expand 1 -fill x
pack .db.frame.dismiss
pack .db.frame

wm transient .db .
wm withdraw  .db

proc popup {message} {
    global message_box
    set message_box "$message"

    # place the dialog box over the control panel, offset just so.

    scan [wm geometry .] "%dx%d+%d+%d" w h x y
    set x [expr "$x + $w/3"]
    set y [expr "$y + $h/3"]
    wm geometry .db "+$x+$y"

    grab set .db
    wm deiconify .db
}
