/*
 * Copyright (c) 1996 Picture Elements, Inc.
 *    Stephen Williams
 *
 *    This source code is free software; you can redistribute it
 *    and/or modify it in source code form under the terms of the GNU
 *    General Public License as published by the Free Software
 *    Foundation; either version 2 of the License, or (at your option)
 *    any later version. In order to redistribute the software in
 *    binary form, you will need a Picture Elements Binary Software
 *    License.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *    You should also have recieved a copy of the Picture Elements
 *    Binary Software License offer along with the source. This offer
 *    allows you to obtain the right to redistribute the software in
 *    binary (compiled) form. If you have not received it, contact
 *    Picture Elements, Inc., 777 Panoramic Way, Berkeley, CA 94704.
 */
#if !defined(WINNT)
#ident "$Id: ucr.c,v 1.59 1998/05/25 00:13:52 steve Exp $"
#endif

/*
 * The communication with the target is, from a linux perspective,
 * very simple. The device only needs to support 2 out and 1 in
 * mailbox, and an in and out bell register set. Since the driver is
 * so amazingly simple, it should work for most PCI interface chips
 * that have uCR on the other side.
 *
 * This driver supports all the channels that the uCR protocol
 * supports, although it assumes that there is at most one open file
 * descriptor per channel. This driver in fact enforces this
 * restriction on the linux side.
 *
 * The driver keeps a pool of input buffers for each open channel, and
 * keeps the target aware of all the pool buffers. The target may DMA
 * packets into the pool buffers in any order without waiting for each
 * packet to be replenished, so burst transfers to the host work as
 * fast as the target can switch DMA values. This driver will attempt
 * to replenish buffers in the pool up to the POOL_BUFFER_COUNT
 * maximum, so it is always ready for a burst.
 *
 * The buffer replenishment scheme allows more buffers to be created
 * to match increased demand, but leaves open the possibility of
 * packets being lost to congested channels. However, If I were to not
 * replenish and drop packets, the congestion would instead become
 * deadlock in the target, and smoothly flowing channels would suffer
 * the deadlock in spite of their good behavior.
 *
 * Reads are interruptible. If I receive a signal while blocked for a
 * read, return a partial read count. If I was unable to transfer any
 * bytes at all, return -EINTR to tell the caller that the read gave up.
 *
 * TERMINOLOGY
 *
 * The packet protocol description define the words read and write from
 * the perspective of the target board. Therefore, the comments in
 * this driver will use the words "in" and "out" to be from the
 * perspective of the host, this driver in particular.
 *
 * Functions that start with dev_ are target specific are are defined
 * in the t-*.h file.
 *
 * HOW IT WORKS
 *
 * The driver keeps a FIFO of commands on the way to the board, and a
 * pool of buffers available to be filled by the target. The command
 * FIFO is called the ``dma_fifo'' in the instance structure and holds
 * a list of command words on their way to the board. The front of the
 * FIFO holds the command currently being transmitted, including the
 * bell mask for telling the control which bells to use to signal the
 * board. This allows attention signals as well as DMA command words
 * to be sent through the same fifo, a reasonable design as
 * the two use the same hardware resources.
 *
 * - THE FIFO
 *
 * The fifo_send() function causes the information to be placed in the
 * fifo, and is called by various parts of the driver. This is the
 * workhorse of the fifo.
 *
 * the fifo_send_done() is called by the interrupt handler to mark
 * that the front item on the fifo has been received by the board. If
 * there is a sync pointer saved, then call wake_up on that pointer. If
 * there are more items, another send is started.
 *
 * The send_dma_values() and send_attention() functions are then
 * convenience functions that take interesting parameters and form
 * them up into the uCR tokens and calls the fifo_send() to get them
 * properly delivered to the board.
 *
 * - THE CHANNEL IN QUEUE
 *
 * Each channel has its own in_queue. This holds buffers filled with
 * data from the board, or waiting for data from the board. The
 * in_ptr and in_nptr members identify the buffers that have been
 * transferred by the target, and the in_sz contains the size filled
 * in (which may be less then PACKET_MTU.)
 *
 * The read_busy and read_nbusy members identify the buffers that are
 * given over to the board but are not yet necessarily
 * transferred. The interrupt handler will move these to the _ptr region
 * as interrupts come in.
 *
 * It is always true that:
 *    (in_ptr + in_nptr) % CHANNEL_BUFS == in_busy
 *    in_nptr + in_nbusy <= CHANNEL_BUFS
 *
 * The in_off member points to the location within the in_ptr
 * buffer and is the next byte to read from the current buffer. It is
 * always true that:
 *    (in_off < in_sz[in_ptr]) || (in_nptr == 0 && in_off == 0)
 *
 * - THE CHANNEL OUT QUEUE
 *
 * Each channel also has an out queue much like the in queue. This
 * is a list of buffers that are ready to be filled by the xxwrite
 * entry, or waiting to be emptied by the board.
 *
 */

# include  "ucrif.h"

/*
 * The details of wiggling the device bits lives here, in the form of
 * macros and inline functions. The os.h defines the interface to the
 * operating system, and t-config.h the interface to the target board.
 */
# include  "os.h"
# include  "t-config.h"
# include  "ucrpriv.h"

#ifndef VENDOR_ID
# error  "The target configuration must define VENDOR_ID"
#endif

#ifndef DEVICE_ID
# error  "The target configuration must define DEVICE_ID"
#endif

#ifndef DEVICE_NAME
# define DEVICE_NAME "uCR"
#endif


unsigned debug_flag = 0;

# define DMA_OUT_BELL_MASK 0x01
# define DMA_IN_BELL_MASK  0x02
# define STATUS_BELL_MASK 0x04
# define ATTEN_BELL_MASK  0x08

enum InlineCmdCode { CMD_NOOP=0, CMD_FILE_MARK=1, CMD_STARTFRAME=2,
		     CMD_STOPFRAME=3 };

const unsigned short vendor_id = VENDOR_ID, device_id = DEVICE_ID;

/*
 * Construct a fresh instance structure. This does not depend on the
 * initial zero value that the C compiler would assign.
 */
void ucr_init_instance(struct Instance*xsp)
{
      unsigned idx;
      xsp->dev = 0;
      xsp->channels = 0;
      xsp->send_busy_flag = 0;
      xsp->send_busy_sync = 0;
      xsp->fifo_ptr = 0;
      xsp->fifo_fill = 0;

      for (idx = 0 ;  idx < 16 ;  idx += 1) {
	    xsp->frame[idx].npages = 0;
	    xsp->frame[idx].pages = 0;
      }
}

void ucr_clear_instance(struct Instance*xsp)
{
      unsigned idx;
      for (idx = 0 ;  idx < 16 ;  idx += 1)
	    if (xsp->frame[idx].npages > 0)
		  ucr_free_frame(xsp, idx);
}

/*
 * This function finds the ChannelData structure that has the
 * specified id. If no structures match, return 0. The result comes
 * from the instance structure, which has a table of opened channels.
 */
static struct ChannelData* channel_by_id(struct Instance*xsp,
					 unsigned short id)
{
      struct ChannelData*cp;

      if (xsp->channels == 0) return 0;
      cp = xsp->channels;

      do {
	    cp = cp->next;
	    if (cp->channel == id) return cp;
      } while (cp != xsp->channels);

      return 0;
}

/*
 * These methods manage the delivery of DMA value sets and ATTENTION
 * signals to the target. Everything happens asynchronously, so this
 * just puts things into fifo queues and lets the interrupt handler
 * drain the fifo as needed.
 *
 * NOTE: The fifo_send may be called in a process context or interrupt
 * handler, but the fifo_send_done is always called by the interrupt
 * handler.
 */

static void write_fifo_values(struct Instance*xsp)
{
	/* Load up the DMA values and interrupt the target. The
	   target will later clear the BELL bit that I set here,
	   and will set the matching IN bit. The interrupt
	   handler will then clear the bit and the
	   send_busy_flag. */

      if (debug_flag & UCR_TRACE_PROTO)
	    printk(DEVICE_NAME "%u (d): set fifo values "
		   "(0x%x, 0x%x, 0x%x)\n",
		   xsp->number,
		   xsp->dma_fifo[xsp->fifo_ptr].dma_addr,
		   xsp->dma_fifo[xsp->fifo_ptr].dma_count,
		   xsp->dma_fifo[xsp->fifo_ptr].bell_mask);

      dev_set_dma_addr(xsp->dev, xsp->dma_fifo[xsp->fifo_ptr].dma_addr);
      dev_set_dma_count(xsp->dev, xsp->dma_fifo[xsp->fifo_ptr].dma_count);
      dev_set_bells(xsp->dev, xsp->dma_fifo[xsp->fifo_ptr].bell_mask);
}

static void fifo_send(struct Instance*xsp, __u32 dma_addr,
		      __u32 dma_count, __u32 mask,
		      struct wait_queue**sync)
{
      unsigned long flags;
      unsigned idx = (xsp->fifo_ptr + xsp->fifo_fill) % FIFO_BUFFER_COUNT;

      if (xsp->fifo_fill >= FIFO_BUFFER_COUNT) {
	    printk(DEVICE_NAME "%u: ERROR: fifo overflow, fifo_fill=%u\n",
		   xsp->number, xsp->fifo_fill);
      }

      flags = dev_mask_irqs(xsp->dev);

      xsp->dma_fifo[idx].dma_addr = dma_addr;
      xsp->dma_fifo[idx].dma_count = dma_count;
      xsp->dma_fifo[idx].bell_mask = mask;
      xsp->dma_fifo[idx].sync   = sync;
      xsp->fifo_fill += 1;

      if (! xsp->send_busy_flag) {

	    xsp->send_busy_flag = 1;
	    write_fifo_values(xsp);
      }

      dev_unmask_irqs(xsp->dev, flags);
}

static void fifo_send_done(struct Instance*xsp)
{
      if (debug_flag & UCR_TRACE_PROTO)
	    printk(DEVICE_NAME "%u (d): done fifo values sync=%p\n",
		   xsp->number, xsp->dma_fifo[xsp->fifo_ptr].sync);

      if (xsp->dma_fifo[xsp->fifo_ptr].sync)
	    wake_up(xsp->dma_fifo[xsp->fifo_ptr].sync);

      xsp->fifo_fill -= 1;
      xsp->fifo_ptr = (xsp->fifo_ptr + 1) % FIFO_BUFFER_COUNT;

      if (xsp->fifo_fill > 0) {
	    write_fifo_values(xsp);
      } else {
	    xsp->send_busy_flag = 0;
	    wake_up(&xsp->send_busy_sync);

	    if (debug_flag & UCR_TRACE_PROTO)
		  printk(DEVICE_NAME "%u (d): set fifo values stopped\n",
			 xsp->number);

      }
}


/*
 * Send a DMA message. This tells the target about a buffer on the
 * host that is available for transfer in whatever direction. The
 * transfer count must not be zero.
 */
static void send_dma_in_values(struct Instance*xsp, __u32 bus_ptr,
			       unsigned long count,
			       struct ChannelData*xpd)
{
      __u32 dma_count = count | (xpd->channel<<16);
      fifo_send(xsp, bus_ptr, dma_count, DMA_IN_BELL_MASK, 0);
      xpd->in_nbusy += 1;
}

static void send_dma_out_values(struct Instance*xsp, __u32 bus_ptr,
				unsigned long count,
				struct ChannelData*xpd)
{
      __u32 dma_count = count | (xpd->channel<<16);
      fifo_send(xsp, bus_ptr, dma_count, DMA_OUT_BELL_MASK, 0);
      xpd->out_nbusy += 1;
}

/*
 * Send an inline command through the channel. An inline command is
 * recognized by a transfer count of 0. In this case, the address is
 * interpreted as the command code.
 */
static void send_inline_out_cmd(struct Instance*xsp, struct ChannelData*xpd,
				enum InlineCmdCode code)
{
      __u32 dma_count = xpd->channel << 16;
      __u32 dma_addr  = code;
      fifo_send(xsp, dma_addr, dma_count, DMA_OUT_BELL_MASK, 0);
}

static __inline__ void send_attention(struct Instance*xsp, unsigned short chan,
				      unsigned short arg1, __u32 arg2,
				      struct wait_queue**sync)
{
      __u32 word1 = chan | (arg1 << 16);

      if (debug_flag & UCR_TRACE_CHAN)
	    printk(DEVICE_NAME
		   "%u.%u (d): Send attention arg1=%u arg2=0x%x\n",
		   xsp->number, chan, arg1, arg2);

      fifo_send(xsp, word1, arg2, ATTEN_BELL_MASK, sync);
}

/*
 * This function sends to the channel all the IN buffers that are not
 * otherwise accounted for. This needs to be called periodically,
 * especially before blocking on a read, to keep the channel flush
 * with buffers it can use.
 */

void ucr_pad_channel(struct Instance*xsp, struct ChannelData*xpd)
{
      unsigned long flags = dev_mask_irqs(xsp->dev);

	/* Here is a good place to do a sanity check, because I'm
	   probably going to be put to sleep right after. */

      if ( (xpd->in_ptr + xpd->in_nptr) % CHANNEL_BUFS !=
	   xpd->in_busy ) {
	    printk(DEVICE_NAME "%u ERROR: Lost sync of read buffers!\n",
		   xsp->number);
      }


      while (xpd->in_nptr + xpd->in_nbusy < CHANNEL_BUFS) {
	    unsigned cur = (xpd->in_busy + xpd->in_nbusy) % CHANNEL_BUFS;
	    send_dma_in_values(xsp, xpd->in_bufs[cur].bus, PACKET_MTU, xpd);
      }

      dev_unmask_irqs(xsp->dev, flags);
}


/*
 * This function causes data in the write buffers to be flushed to the
 * channel. It will wait until the target reads the data and sends a
 * completion status, so when this function ends, be assured that the
 * channel data has been read.
 */
static void flush_channel(struct Instance*xsp, struct ChannelData*xpd)
{
      unsigned long mask;

      if (debug_flag&UCR_TRACE_CHAN)
	    printk(DEVICE_NAME "%u.%u (d): Flushing write buffer "
		   "count=%u, out_ptr=%u, out_nbusy=%u\n",
		   xsp->number, xpd->channel, xpd->out_off,
		   xpd->out_ptr, xpd->out_nbusy);

      mask = dev_mask_irqs(xsp->dev);

	/* Flush the last write buffer, if there are any bytes that
	   have not been transmitted. Then wait for all the
	   outstanding writes to complete to the board. */

      if (xpd->out_off > 0) {
	    __u32 buf = xpd->out_bufs[xpd->out_ptr].bus;
	    send_dma_out_values(xsp, buf, xpd->out_off, xpd);
      }

      xpd->out_ptr = (xpd->out_ptr + 1) % CHANNEL_BUFS;
      xpd->out_off = 0;

      while (xpd->out_nbusy > 0) {

	    if (debug_flag&UCR_TRACE_CHAN)
		  printk(DEVICE_NAME "%u.%u (d): %u buffers to go...\n",
			 xsp->number, xpd->channel, xpd->out_nbusy);

	    if (is_signalled(xsp)) {
		  if (debug_flag & UCR_TRACE_CHAN)
			printk(DEVICE_NAME "%u.%u (d): "
			       "flush_channel interrupted "
			       "(%u busy buffers)\n", xsp->number,
			       xpd->channel, xpd->out_nbusy);
		  break;
	    }

	    atomic_sleep_on(xsp, mask, &xpd->out_sync);
      }

      dev_unmask_irqs(xsp->dev, mask);

      if (debug_flag&UCR_TRACE_CHAN)
	    printk(DEVICE_NAME "%u.%u (d): Flushing done.\n",
		   xsp->number, xpd->channel);
}


/*
 * There is some shared channel state between host and target that
 * needs to be cleared before a host file can disconnect from the
 * channel. This function sends a reset to the target channel to clear
 * that state. When this function returns, it is safe to drop the
 * channel.
 */
static void reset_channel(struct Instance*xsp, struct ChannelData*xpd)
{
      struct wait_queue*release_sync = 0;
      unsigned long mask;

      if (debug_flag&UCR_TRACE_CHAN)
	    printk(DEVICE_NAME "%u.%u (d): channel reset\n",
		   xsp->number, xpd->channel);

	/* Reset the channel on the target by sending the reset
	   attention signal. */

      mask = dev_mask_irqs(xsp->dev);
      if (is_signalled(xsp)) {
	    send_attention(xsp, xpd->channel, 0, 1, 0);
      } else {
	    send_attention(xsp, xpd->channel, 0, 1, &release_sync);
	    atomic_sleep_on(xsp, mask, &release_sync);
      }
      dev_unmask_irqs(xsp->dev, mask);

      if (is_signalled(xsp)) {
	    printk(DEVICE_NAME "%u WARNING: channel %u interrupted!\n",
		   xsp->number, xpd->channel);
      }
}

static int ucr_send_frame(struct Instance*xsp, struct ChannelData*xpd,
			  unsigned id)
{
      __u32 bus, *buf;
      unsigned cur;
      unsigned long mask;

      flush_channel(xsp, xpd);
      if (is_signalled(xsp)) return -EINTR;

      if (debug_flag & UCR_TRACE_CHAN)
	    printk(DEVICE_NAME "%u (d): FRAME %u -- Start sending frame."
		   " out_nbusy=%u\n", xsp->number, id, xpd->out_nbusy);


	/* Write the FRAME header block is a single transfer. This is
	   required by the packet protocol for sending frames. We know
	   that there is a buffer available and completely empty by
	   the flush_channel above. */

      buf = xpd->out_bufs[xpd->out_ptr].host;
      bus = xpd->out_bufs[xpd->out_ptr].bus;
      buf[0] = xsp->frame[id].pagesize * xsp->frame[id].npages;
      buf[1] = xsp->frame[id].pagesize;
      buf[2] = xsp->frame[id].npages;
      buf[3] = 0;

      if (debug_flag & UCR_TRACE_CHAN)
	    printk(DEVICE_NAME "%u (d): FRAME %u -- Send frame header "
		   "(size=%u, page=%u, npages=%u) out_nbusy=%u\n",
		   xsp->number, id, buf[0], buf[1], buf[2], xpd->out_nbusy);

      mask = dev_mask_irqs(xsp->dev);
      send_inline_out_cmd(xsp, xpd, CMD_STARTFRAME);
      send_dma_out_values(xsp, bus, 4*sizeof(buf[0]), xpd);
      STEP_BUF_PTR(xpd->out_ptr);
      dev_unmask_irqs(xsp->dev, mask);

      cur = 0;
      while (cur < xsp->frame[id].npages) {
	    unsigned trans;
	    unsigned idx;

	    mask = dev_mask_irqs(xsp->dev);
	    while (xpd->out_nbusy == CHANNEL_BUFS) {
		  atomic_sleep_on(xsp, mask, &xpd->out_sync);

		  if (is_signalled(xsp)) {
			dev_unmask_irqs(xsp->dev, mask);
			goto signalled;
		  }
	    }
	    dev_unmask_irqs(xsp->dev, mask);

	    buf = xpd->out_bufs[xpd->out_ptr].host;
	    bus = xpd->out_bufs[xpd->out_ptr].bus;

	    for (idx = cur, trans = 0
		       ; (idx < xsp->frame[id].npages) && (trans < PACKET_MTU)
		       ; idx += 1, trans += 4) {
		  buf[idx-cur] = xsp->frame[id].pages[idx].bus;
	    }

	    if (debug_flag & UCR_TRACE_CHAN)
		  printk(DEVICE_NAME "%u (d): FRAME %u -- Send frame data "
			 "(npages=%u, trans=%u)\n", xsp->number, id,
			 idx - cur, trans);

	    mask = dev_mask_irqs(xsp->dev);
	    send_dma_out_values(xsp, bus, trans, xpd);
	    STEP_BUF_PTR(xpd->out_ptr);
	    dev_unmask_irqs(xsp->dev, mask);

	    cur = idx;
      }

      mask = dev_mask_irqs(xsp->dev);
      send_inline_out_cmd(xsp, xpd, CMD_STOPFRAME);
      dev_unmask_irqs(xsp->dev, mask);

      if (debug_flag & UCR_TRACE_CHAN)
	    printk(DEVICE_NAME "%u (d): FRAME %u -- done sending frame."
		   " out_nbusy=%u\n", xsp->number, id, xpd->out_nbusy);

      return 0;

signalled:
      if (debug_flag & UCR_TRACE_CHAN)
	    printk(DEVICE_NAME "%u (d): FRAME %u -- signalled during send "
		   "frame!\n", xsp->number, id);

      return -EINTR;
}

static long ucr_channel_nread(struct Instance*xsp, struct ChannelData*xpd)
{
      long count = 0;
      if (xpd->in_nptr == 0)
	    return 0;

	/* For now, just look at the size of the first buffer. There
	   may be more, but I never did make any guarantees. */
      count += xpd->in_sz[xpd->in_ptr] - xpd->in_off;


      if (debug_flag & UCR_TRACE_CHAN)
	    printk(DEVICE_NAME "%u.%u (d): NREAD returns %u\n",
		   xsp->number, xpd->channel, count);

      return count;
}

/*
 * When a new file is opened, create a ChannelData structure to be
 * associated with that descriptor. The file by default gets channel
 * 0, but the application must somehow assure that there are no two
 * files opened to the same channel, once real I/O occurs.
 */
int ucr_open(struct Instance*xsp, struct ChannelData*xpd)
{
      unsigned idx;

	/* Prevent a duplicate open of channel 0. */

      if (channel_by_id(xsp, 0)) return -EBUSY;
      xpd->channel = 0;

	/* Initialize the write buffers ... */
      xpd->out_busy = 0;
      xpd->out_nbusy = 0;
      xpd->out_ptr = 0;
      xpd->out_off = 0;
      xpd->out_sync = 0;
      xpd->out_status_sync = 0;
      for (idx = 0 ;  idx < CHANNEL_BUFS ;  idx += 1) {
	    xpd->out_bufs[idx] = alloc_buf();
      }

	/* initialize the in buffers... */
      xpd->in_busy = 0;
      xpd->in_nbusy = 0;
      xpd->in_ptr = 0;
      xpd->in_nptr = 0;
      xpd->in_off = 0;
      xpd->in_sync = 0;
      for (idx = 0 ;  idx < CHANNEL_BUFS ;  idx += 1) {
	    xpd->in_bufs[idx] = alloc_buf();
	    xpd->in_sz[idx] = 0;
      }

	/* Reset the channel on the target, so I know I have a
	   reasonably clean slate. I do not need to wait for the reset
	   to complete here, so do not use the regular function. */

      send_attention(xsp, xpd->channel, 0, 1, 0);


	/* Put the channel information into the channel list so that
	   arriving packets can be properly dispatched to the
	   process. */

      if (xsp->channels == 0) {
	    xsp->channels = xpd;
	    xpd->next = xpd;
	    xpd->prev = xpd;
      } else {
	    xpd->next = xsp->channels;
	    xpd->prev = xsp->channels->prev;
	    xpd->next->prev = xpd;
	    xpd->prev->next = xpd;
      }

      return 0;
}

/*
 * Release the file. Take the ChannelData structure out of the channel
 * list and return all the unread buffers to the pool, for use by
 * other files. (put_pool() will delete excess buffers.)
 */

void ucr_release(struct Instance*xsp, struct ChannelData*xpd)
{
      unsigned idx;

      flush_channel(xsp, xpd);

      reset_channel(xsp, xpd);


      if (debug_flag&UCR_TRACE_CHAN)
	    printk(DEVICE_NAME "%u.%u (d): releasing buffers\n",
		   xsp->number, xpd->channel);

      for (idx = 0 ;  idx < CHANNEL_BUFS ;  idx += 1)
	    free_buf(xpd->out_bufs[idx]);


	/* No more communication with the target channel, so remove
	   the ChannelData structure from the channel list, remove
	   buffers, and release the xpd object. */

      if (xsp->channels == xpd) xsp->channels = xsp->channels->next;
      if (xsp->channels == xpd) {
	    xsp->channels = 0;
      } else {
	    xpd->prev->next = xpd->next;
	    xpd->next->prev = xpd->prev;
      }


	/* Clean up the read buffers */

      for (idx = 0 ;  idx < CHANNEL_BUFS ;  idx += 1)
	    free_buf(xpd->in_bufs[idx]);


      if (debug_flag&UCR_TRACE_CHAN)
	    printk(DEVICE_NAME "%u.%u (d): channel closed.\n",
		   xsp->number, xpd->channel);

}

/*
 * Read works by waiting for at least one packet to be dispatched to
 * the channel. I then copy the bytes out of the packet into the
 * reader's buffer, and as I drain packets return them to the read
 * pool.
 *
 * If the block_flag is true, then this read will block until the
 * requested count is read. Otherwise, do not block.
 */
long ucr_read(struct Instance*xsp, struct ChannelData*xpd,
	      char *bytes, unsigned long count, int block_flag)
{
      unsigned tcount = count;

      if (debug_flag & UCR_TRACE_CHAN)
	    printk(DEVICE_NAME
		   "%u.%u (d): ucr_read %u bytes, block=%s\n",
		   xsp->number, xpd->channel, count,
		   block_flag?"true":"false");


      while (tcount > 0) {

	    unsigned long mask;
	    unsigned trans;
	    void*buf;

	    if (xpd->in_nptr == 0) {

		  ucr_pad_channel(xsp, xpd);
		  flush_channel(xsp, xpd);

		    /* If I transferred at least a few bytes to the
		       caller, then that is good enough. Don't block. */
		  if (tcount < count)
			break;

		    /* I'm about to block. If the caller doesn't want
		       that, then break from the read loop here. */
		  if (!block_flag)
			break;

		  if (wait_for_read_data(xsp, xpd) == -EINTR)
			goto signalled;

	    }

	    buf = xpd->in_bufs[xpd->in_ptr].host;

	    trans = tcount;
	    if ((trans + xpd->in_off) > xpd->in_sz[xpd->in_ptr])
		  trans = xpd->in_sz[xpd->in_ptr] - xpd->in_off;

	    if (debug_flag & UCR_TRACE_CHAN)
		  printk(DEVICE_NAME
			 "%u.%u (d): read %u of %u bytes "
			 "[in_off=%u, in_ptr=%u]\n", xsp->number,
			 xpd->channel, trans, xpd->in_sz[xpd->in_ptr],
			 xpd->in_off, xpd->in_ptr);

	    copy_to_user(bytes, (char*)buf + xpd->in_off, trans);

	    tcount -= trans;
	    bytes += trans;
	    xpd->in_off += trans;

	      /* If I get to the end of a buffer, immediately send it
		 back to the target for refilling. Do it this way
		 because I may be able to prevent a sleep for large
		 reads. */
	    if (xpd->in_off == xpd->in_sz[xpd->in_ptr]) {
		  mask = dev_mask_irqs(xsp->dev);
		  send_dma_in_values(xsp, xpd->in_bufs[xpd->in_ptr].bus,
				     PACKET_MTU, xpd);
		  STEP_BUF_PTR(xpd->in_ptr);
		  xpd->in_nptr -= 1;
		  xpd->in_off = 0;
		  dev_unmask_irqs(xsp->dev, mask);
	    }
      }

	/* Calculate the return value. */
      count = count - tcount;

      if (debug_flag & UCR_TRACE_CHAN)
	    printk(DEVICE_NAME "%u.%u (d): ucr_read complete %d bytes\n",
		   xsp->number, xpd->channel, count);


      return count;

signalled:
      if (debug_flag & UCR_TRACE_CHAN)
	    printk(DEVICE_NAME "%u.%u (d): ucr_read interrupted "
		   "after %u bytes\n", xsp->number, xpd->channel,
		   count - tcount);

      if (count == tcount) return -EINTR;
      else return count - tcount;
}


/*
 * The write works by waiting for the previous DMA write to finish,
 * then starting a new one for this data. The DMA operation may
 * proceed without me waiting around in the write, but I can't start a
 * new one until the DMA completes. For this to work, the mark bit 0
 * must start out as 1.
 */

long ucr_write(struct Instance*xsp, struct ChannelData*xpd,
	       const char*bytes, const unsigned long count)
{
      int rc;
      unsigned tcount = count;

      if (debug_flag & UCR_TRACE_CHAN)
	    printk(DEVICE_NAME "%u.%u (d): ucr_write %u bytes\n",
		   xsp->number, xpd->channel, count);

      while (tcount > 0) {
	    unsigned trans = tcount;
	    void*buf;
	    __u32 bus;

	      /* Wait for some buffer space to free up. */
	    if (wait_for_write_space(xsp, xpd) == -EINTR)
		  goto signalled;

	    buf = xpd->out_bufs[xpd->out_ptr].host;
	    bus = xpd->out_bufs[xpd->out_ptr].bus;

	    if ((xpd->out_off + trans) > PACKET_MTU)
		  trans = PACKET_MTU - xpd->out_off;

	    if (debug_flag & UCR_TRACE_CHAN)
		  printk(DEVICE_NAME "%u.%u (d): write %u bytes to out_ptr=%u"
			 "(buf=0x%x) offset=%u\n", xsp->number,
			 xpd->channel, trans, xpd->out_ptr,
			 xpd->out_bufs[xpd->out_ptr].host, xpd->out_off);

	    copy_from_user((char*)buf + xpd->out_off, bytes, trans);

	    xpd->out_off += trans;
	    tcount -= trans;
	    bytes  += trans;

	    if (xpd->out_off == PACKET_MTU) {
		  const unsigned long mask = dev_mask_irqs(xsp->dev);
		  send_dma_out_values(xsp, bus, xpd->out_off, xpd);
		  dev_unmask_irqs(xsp->dev, mask);

		  STEP_BUF_PTR(xpd->out_ptr);
		  xpd->out_off = 0;
	    }
      }


      if (debug_flag & UCR_TRACE_CHAN)
	    printk(DEVICE_NAME "%u.%u (d): ucr_write complete %d bytes\n",
		   xsp->number, xpd->channel, count);

	/* I expect that a write may cause the target board to send
	   data back. At any rate, the target board may be blocking on
	   a write (to me!) so will need some loose buffers before it
	   can read what I'm writing. */
      ucr_pad_channel(xsp, xpd);

      return count;


signalled:
	/* If waiting for buffer space for the output is interrupted
	   by a signal, exit here. */

      if (debug_flag & UCR_TRACE_CHAN)
	    printk(DEVICE_NAME "%u.%u (d): ucr_write interrupted\n",
		   xsp->number, xpd->channel);

      if (count == tcount) return -EINTR;
      else return count - tcount;
}

int ucr_ioctl(struct Instance*xsp, struct ChannelData*xpd,
	      unsigned int cmd, unsigned long arg)
{
      int rc;

      switch (cmd) {

	  case UCR_FLUSH:
	    flush_channel(xsp, xpd);
	    return 0;

	  case UCR_CHANNEL:
	    if (debug_flag & UCR_TRACE_CHAN)
		  printk(DEVICE_NAME "%u.%u (d): switch to channel %u\n",
			 xsp->number, xpd->channel, arg);

	    if (arg >= 0xffff) return -EINVAL;
	    if (xpd->channel == arg) return 0;
	    if (channel_by_id(xsp, (__u16)arg)) return -EBUSY;
	    flush_channel(xsp, xpd);
	    reset_channel(xsp, xpd);
	    xpd->channel = (__u16)arg;

	    if (debug_flag & UCR_TRACE_CHAN)
		  printk(DEVICE_NAME "%u.%u (d): switch complete\n",
			 xsp->number, xpd->channel);

	    return 0;

	  case UCR_ATTENTION:
	    if (debug_flag & UCR_TRACE_CHAN)
		  printk(DEVICE_NAME "%u.%u (d): software attention %u\n",
			 xsp->number, xpd->channel, arg);

	    send_attention(xsp, xpd->channel, 1, arg, 0);
	    return 0;

	  case UCR_SEND_FILE_MARK:
	    if (debug_flag & UCR_TRACE_CHAN)
		  printk(DEVICE_NAME "%u.%u (d): send file mark\n",
			 xsp->number, xpd->channel);

	    flush_channel(xsp, xpd);
	    {
		  unsigned long flags = dev_mask_irqs(xsp->dev);
		  send_inline_out_cmd(xsp, xpd, CMD_FILE_MARK);
		  atomic_sleep_on(xsp, flags, &xpd->out_status_sync);
		  dev_unmask_irqs(xsp->dev, flags);
	    }

	    return 0;

	  case UCR_MAKE_FRAME: {
		unsigned id = 0xf & (arg >> 28);
		return ucr_make_frame(xsp, id, arg&0x0fffffff);
	  }

	  case UCR_FREE_FRAME: {
		unsigned id = 0xf & (arg >> 28);
		return ucr_free_frame(xsp, id);
	  }

	  case UCR_SEND_FRAME: {
		unsigned id = 0xf & (arg >> 28);
		return ucr_send_frame(xsp, xpd, id);
	  }

	  case UCR_NREAD: {
		return ucr_channel_nread(xsp, xpd);
	  }
      }

      return -ENOTTY;
}

/*
 * This function handles the status words that come from the
 * target. Mostly, they are mark commands.
 */

static void status_irq(struct Instance*xsp, __u16 cmd, __u16 arg1, __u32 arg2)
{
      struct ChannelData*xpd;

      switch (cmd) {

	      /* This is a DMA read value set completion, meaning a
		 DMA_OUT has been finished.

		 arg1 == channel
		 arg2 == buffer address */

	  case 0:
	    xpd = channel_by_id(xsp, arg1);
	    if (xpd == 0) {
		  printk(DEVICE_NAME "%u ERROR: OUT STATUS for missing "
			 "channel %u\n", xsp->number, arg1);
		  break;
	    }

	    if (debug_flag & UCR_TRACE_PROTO)
		  printk(DEVICE_NAME "%u.%u (d): Got OUT completion "
			 "with %u pending: ptr=0x%x\n",
			 xsp->number, arg1, xpd->out_nbusy, arg2);


	    xpd->out_busy = (xpd->out_busy + 1) % CHANNEL_BUFS;
	    xpd->out_nbusy -= 1;
	    wake_up(&xpd->out_sync);

	    break;

	      /* This is a DMA write value set completion, meaning a
		 DMA_IN has been finished. The target board returns in
		 arg2 the pointer plus the number of bytes
		 transmitted, so I can use that to get the sz and make
		 sure things are in sync.

		 arg1 == channel
		 arg2 == buffer address + transfer count */

	  case 1:

	    xpd = channel_by_id(xsp, arg1);
	    if (xpd == 0) {
		  printk(DEVICE_NAME "%u ERROR: IN STATUS for missing "
			 "channel %u\n", xsp->number, arg1);
		  break;
	    }

	    xpd->in_sz[xpd->in_busy] = arg2 - xpd->in_bufs[xpd->in_busy].bus;

	    if (debug_flag & UCR_TRACE_PROTO)
		  printk(DEVICE_NAME "%u.%u (d): Got IN completion "
			 "with %u pending: ptr=0x%x (base=0x%x, count=%u)\n",
			 xsp->number, xpd->channel, xpd->in_nbusy, arg2,
			 xpd->in_bufs[xpd->in_busy].bus,
			 xpd->in_sz[xpd->in_busy]);

	    xpd->in_busy = (xpd->in_busy + 1) % CHANNEL_BUFS;
	    xpd->in_nbusy -= 1;
	    xpd->in_nptr += 1;
	    wake_up(&xpd->in_sync);

	    break;


	      /* DMA read values inline command complete.

		 arg1 == channel
		 arg2 == return of the cookie and command bits. */

	  case 2:

	    xpd = channel_by_id(xsp, arg1);
	    if (xpd == 0) {
		  printk(DEVICE_NAME "%u ERROR: IN STATUS for missing "
			 "channel %u\n", xsp->number, arg1);
		  break;
	    }
	    if (debug_flag & UCR_TRACE_PROTO)
		  printk(DEVICE_NAME "%u.%u (d): Got OUT cmd status "
			 "cmd=0x%x\n",
			 xsp->number, xpd->channel, arg2);
	    wake_up(&xpd->out_status_sync);
	    break;

	  case 3:
	  default:
	    printk(DEVICE_NAME "%u: funky STATUS %u(0x%x, 0x%x)\n",
		   xsp->number, cmd, arg1, arg2);
	    break;
      }
}


void ucr_irq(struct Instance*xsp)
{
      unsigned long mask;

	/* Notice any bells that might be happening and clear the
	   interrupt flags for them. */

      if ( (mask = dev_get_bells(xsp->dev)) ) {
	    unsigned idx;

	    if (debug_flag  & UCR_TRACE_PROTO)
		  printk(DEVICE_NAME "%u (d): Got bells mask=0x%x\n",
			 xsp->number, mask);


	      /* Bell 0 is notice that the DMA value set registers is
		 available. Clear the busy flag and awaken a thread
		 waiting for this occasion.

		 The same if the registers have been used to send an
		 attention. The only difference in that case is that
		 Bell 2 was used to interrupt the target. */

	    if (mask & (DMA_IN_BELL_MASK|ATTEN_BELL_MASK|DMA_OUT_BELL_MASK)) {
		  fifo_send_done(xsp);
	    }

	      /* Bell 1 is notice that the board is sending me a
		 status word. The status word is a set of bits that
		 reflect the current status of things, and is resent
		 whenever the status changes. Read it and process it. */

	    if (mask & STATUS_BELL_MASK) {

		  __u32 status = dev_get_status(xsp->dev);
		  __u32 arg = dev_get_status_arg(xsp->dev);

		  dev_set_bells(xsp->dev, STATUS_BELL_MASK);
		  status_irq(xsp, (__u16)(status&0xffff),
			     (__u16)(status>>16), arg);
	    }

      }

}

/*
 * $Log: ucr.c,v $
 * Revision 1.59  1998/05/25 00:13:52  steve
 *  fix an abort panic.
 *
 * Revision 1.58  1998/04/30 01:57:09  steve
 *  Canonical messages.
 *
 * Revision 1.57  1998/04/22 19:06:57  steve
 *  Some write cancel support.
 *
 * Revision 1.56  1998/04/18 02:43:54  steve
 *  Fix trouble removing device lines at unload.
 *  Add support for read cancellation under NT.
 *  isecons device (nt only) now throws away old
 *  data instead of truncating new.
 *
 * Revision 1.55  1998/04/07 22:06:11  steve
 *  More tracing support.
 *
 * Revision 1.54  1998/04/07 18:43:50  steve
 *  Report channel number consistently.
 *
 * Revision 1.53  1998/04/01 01:11:19  steve
 *  commentary typo.
 *
 * Revision 1.52  1998/04/01 00:23:05  steve
 *  Support partial reads.
 *
 * Revision 1.51  1998/03/23 23:40:21  steve
 *  frames work under NT.
 *
 * Revision 1.50  1998/03/23 03:30:25  steve
 *  non-blocking reading under NT.
 *
 * Revision 1.49  1998/03/22 21:07:17  steve
 *  Fix frame management bugs.
 *
 * Revision 1.48  1998/03/21 21:06:56  steve
 *  Rearrange frame handling code.
 *
 * Revision 1.47  1998/03/12 22:55:30  steve
 *  Fix frame release
 *
 * Revision 1.46  1998/03/11 00:07:17  steve
 *  Support non-blocking input, and add an ioctl
 *  that tests for input data.
 *
 * Revision 1.45  1998/01/29 22:00:06  steve
 *  Flush writes when read blocks.
 *
 * Revision 1.44  1997/12/12 04:29:10  steve
 *  Error linking driver
 *
 * Revision 1.43  1997/12/07 02:16:43  steve
 *  Fix NT blocking problems, redo the console to
 *  not block and keep a limited buffer, trace the
 *  ucrx operations, and ucrx operations work on NT.
 *
 * Revision 1.42  1997/12/02 02:29:04  steve
 *  Add support for ISE board reset.
 *
 * Revision 1.41  1997/11/16 11:37:44  steve
 *  Make NT driver asynchronous.
 *
 * Revision 1.40  1997/11/13 19:21:58  demo
 *  Typo in printk.
 *
 * Revision 1.39  1997/11/11 11:19:46  steve
 *  Add support for frames.
 *
 * Revision 1.38  1997/10/30 19:47:27  steve
 *  properly init status sync.
 *
 * Revision 1.37  1997/09/12 05:18:07  steve
 *  Port to linux 2.1
 *
 * Revision 1.36  1997/09/11 21:58:36  steve
 *  Port the uCR device driver to Windows NT,
 *  make the configure select the target board
 *  and target host interface headers.
 *
 * Revision 1.35  1997/08/26 18:25:32  steve
 *  read/write names to in/out for clarity.
 *
 * Revision 1.34  1997/08/19 07:23:05  steve
 *  Slight performance improvement.
 *
 * Revision 1.33  1997/08/18 21:53:46  steve
 *  Support inline channel commands.
 */
