/*
 * Copyright (c) 1997 Picture Elements, Inc.
 *    Stephen Williams (steve@picturel.com)
 *
 *    This source code is free software; you can redistribute it
 *    and/or modify it in source code form under the terms of the GNU
 *    General Public License as published by the Free Software
 *    Foundation; either version 2 of the License, or (at your option)
 *    any later version. In order to redistribute the software in
 *    binary form, you will need a Picture Elements Binary Software
 *    License.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *  ---
 *    You should also have recieved a copy of the Picture Elements
 *    Binary Software License offer along with the source. This offer
 *    allows you to obtain the right to redistribute the software in
 *    binary (compiled) form. If you have not received it, contact
 *    Picture Elements, Inc., 777 Panoramic Way, Berkeley, CA 94704.
 */
#if !defined(WIN32)
#ident "$Id: Ucr_Init.c,v 1.23 1998/07/11 21:29:04 steve Exp $"
#endif

# include  <tcl.h>
# include  <ucrif.h>
# include  <assert.h>
# include  <stdlib.h>
# include  <malloc.h>
# include  <string.h>
# include  <errno.h>
# include  "priv.h"


static Tcl_HashTable ucr_item_table;
static Tcl_HashTable ucr_frame_table;

static int frame_get_ch(struct ucr_frame_info*finfo, unsigned long off,
			Tcl_Interp*interp, int argc, char*argv[])
{
      Tcl_Channel ch;
      unsigned long len;
      char*buf;
      int mode;

      if (argc < 2) {
	    Tcl_SetResult(interp, "missing channel name", TCL_STATIC);
	    return TCL_ERROR;
      }

      ch = Tcl_GetChannel(interp, argv[1], &mode);
      if ((ch == 0) || (! mode&TCL_WRITABLE)) {
	    Tcl_SetResult(interp, "no such writable channel", TCL_STATIC);
	    return TCL_ERROR;
      }

      if (argc > 2) {
	    char*end;
	    len = strtoul(argv[2],&end,0);
	    if (*end || (off + len) >= finfo->frame_size) {
		  Tcl_AppendResult(interp, "invalid length: ",
				   argv[2], 0);
		  return TCL_ERROR;
	    }

      } else {

	    len = finfo->frame_size - off;
      }

      buf = (char*)finfo->frame_base;

      while (len > 0) {
	    int rc = Tcl_Write(ch, buf, len);
	    if (rc == -1) {
		  Tcl_SetResult(interp, "error writing to channel",
				TCL_STATIC);
		  return TCL_ERROR;
	    }

	    len -= rc;
	    buf += rc;
      }

      return TCL_OK;
}

static int frame_put_ch(struct ucr_frame_info*finfo, unsigned long off,
			Tcl_Interp*interp, int argc, char*argv[])
{
      Tcl_Channel ch;
      unsigned long len;
      char*buf;
      int mode;

      if (argc < 2) {
	    Tcl_SetResult(interp, "missing channel name", TCL_STATIC);
	    return TCL_ERROR;
      }

      ch = Tcl_GetChannel(interp, argv[1], &mode);
      if ((ch == 0) || (! mode&TCL_READABLE)) {
	    Tcl_SetResult(interp, "no such readable channel", TCL_STATIC);
	    return TCL_ERROR;
      }

      if (argc > 2) {
	    char*end;
	    len = strtoul(argv[2],&end,0);
	    if (*end || (off + len) >= finfo->frame_size) {
		  Tcl_SetResult(interp, "invalid length", TCL_STATIC);
		  return TCL_ERROR;
	    }

      } else {
	    len = finfo->frame_size - off;

      }

      buf = (char*)finfo->frame_base;

      while ((len > 0) && (! Tcl_Eof(ch))) {
	    int rc = Tcl_Read(ch, buf, len);
	    if (rc == -1) {
		  Tcl_SetResult(interp, "error writing to channel",
				TCL_STATIC);
		  return TCL_ERROR;
	    }

	    len -= rc;
	    buf += rc;
      }

      return TCL_OK;
}

static int frame_proc(ClientData cd, Tcl_Interp*interp, int argc, char*argv[])
{
      struct ucr_frame_info*finfo = (struct ucr_frame_info*)cd;

      if (argc == 1)
	    return TCL_OK;

      if (strcmp("get",argv[1]) == 0) {
	    unsigned long off;
	    char*end;
	    if (argc < 4) {
		  Tcl_SetResult(interp, "missing get parameters.", TCL_STATIC);
		  return TCL_ERROR;
	    }

	    off = strtoul(argv[2],&end,0);
	    if (*end) {
		  Tcl_SetResult(interp, "invalid offset.", TCL_STATIC);
		  return TCL_ERROR;
	    }

	    if (strcmp("-ch",argv[3]) == 0)
		  return frame_get_ch(finfo, off, interp, argc-3, argv+3);

	    Tcl_SetResult(interp, "invalid frame data source.", TCL_STATIC);
	    return TCL_ERROR;
      }

      if (strcmp("put",argv[1]) == 0) {
	    unsigned long off;
	    char*end;
	    if (argc < 4) {
		  Tcl_SetResult(interp, "missing get parameters.", TCL_STATIC);
		  return TCL_ERROR;
	    }

	    off = strtoul(argv[2],&end,0);
	    if (*end) {
		  Tcl_SetResult(interp, "invalid offset.", TCL_STATIC);
		  return TCL_ERROR;
	    }

	    if (strcmp("-ch",argv[3]) == 0)
		  return frame_put_ch(finfo, off, interp, argc-3, argv+3);

	    Tcl_SetResult(interp, "invalid frame data source.", TCL_STATIC);
	    return TCL_ERROR;
      }

      Tcl_SetResult(interp, "invalid frame subcommand.", TCL_STATIC);
      return TCL_ERROR;
}

static void frame_del(ClientData cd)
{
      struct ucr_frame_info*finfo = (struct ucr_frame_info*)cd;
      struct ucr_channel_info*cinfo = finfo->ch;

      if (cinfo->frames == finfo) {
	    cinfo->frames = finfo->next;

      } else {
	    struct ucr_frame_info*prev = cinfo->frames;
	    while (prev->next != finfo)
		  prev = prev->next;

	    prev->next = finfo->next;
      }

      finfo->ch = 0;
      finfo->next = 0;

      ucr_manage_frame(0, finfo, 0);
      free(finfo);
}

static int ucr_frame(ClientData xx, Tcl_Interp*interp, int argc, char*argv[])
{
      struct ucr_channel_info*cinfo;
      struct ucr_frame_info*finfo;
      Tcl_HashEntry*cep, *fep;
      unsigned long fnum, fsize;
      char*end;
      int new_flag;

      if (argc != 5) {
	    Tcl_SetResult(interp, "usage: ucr_frame <name> <ch> "
			  "<num> <size>", TCL_STATIC);
	    return TCL_ERROR;
      }

      cep = Tcl_FindHashEntry(&ucr_item_table, argv[2]);
      if (cep == 0) {
	    Tcl_SetResult(interp, "not a valid ucr channel.", TCL_STATIC);
	    return TCL_ERROR;
      }

      cinfo = (struct ucr_channel_info*)Tcl_GetHashValue(cep);

      fnum = strtoul(argv[3],&end,0);
      if (*end || (fnum >= 16)) {
	    Tcl_SetResult(interp, "invalid frame number", TCL_STATIC);
	    return TCL_ERROR;
      }

      fsize = strtoul(argv[4],&end,0);
      if (*end || (fsize == 0) || (fsize & 0xf0000000)) {
	    Tcl_SetResult(interp, "invalid frame size.", TCL_STATIC);
	    return TCL_ERROR;
      }

	/* XXXX Check for the frame number busy? */

      fep = Tcl_CreateHashEntry(&ucr_frame_table, argv[1], &new_flag);
      if (new_flag == 0) {
	    Tcl_SetResult(interp, "frame name is busy.", TCL_STATIC);
	    return TCL_ERROR;
      }

      finfo = calloc(1, sizeof *finfo);
      finfo->cinfo = cinfo;
      finfo->num = fnum;
      finfo->next = cinfo->frames;
      cinfo->frames = finfo;

      ucr_manage_frame(interp, finfo, fsize);
      Tcl_CreateCommand(interp, argv[1], frame_proc, finfo, frame_del);

      {
	    char buf[16];
	    sprintf(buf, "%lu", finfo->frame_size);
	    Tcl_SetResult(interp, buf, TCL_VOLATILE);
      }

      return TCL_OK;
}


/*
 * The ucr_read_proc function provides support for asynchronous
 * intput. The read proc is called by the notifier when something
 * interesting happens on the channel.
 */

void ucr_read_proc(ClientData cd, int mask)
{
      struct ucr_channel_info*ci = (struct ucr_channel_info*)cd;

      if (ci->read_cmd && (mask & TCL_READABLE))
	    Tcl_GlobalEval(ci->interp, ci->read_cmd);
}

static int ucrBlockModeProc(ClientData cd, int mode)
{
      struct ucr_channel_info*info = (struct ucr_channel_info*)cd;
      info->block_mode = mode;
      return 0;
}

static int ucrCloseProc(ClientData cd, Tcl_Interp*interp)
{
      struct ucr_channel_info*info = (struct ucr_channel_info*)cd;
      close(info->fd);
      Tcl_DeleteHashEntry(info->self);
      if (info->read_cmd)
	    free(info->read_cmd);
      free(info);
      return 0;
}

/*
 * Seek is not supported for this kind of file. Nor are there any
 * options that are supported.
 */
static int ucrSeekProc(ClientData cd, long offset, int mode, int *err)
{
      *err = EINVAL;
      return -1;
}

static int ucrSetOptionsProc(ClientData cd, Tcl_Interp*interp,
			     char*opt, char*val)
{
      return TCL_ERROR;
}

static int ucrGetOptionsProc(ClientData cd, Tcl_Interp*interp,
			     char*name, Tcl_DString*ds)
{
      Tcl_SetErrno(EINVAL);
      return TCL_ERROR;
}

static int ucrGetHandleProc(ClientData cd, int dir, ClientData *hp)
{
      return TCL_ERROR;
}

static const struct Tcl_ChannelType ucr_channel_type = {
      "ucr-channel",
      ucrBlockModeProc, /* Block Mode Procedure */
      ucrCloseProc, /* Close procedure */
      ucrInputProc, /* input procedure */
      ucrOutputProc, /* output procedure */
      ucrSeekProc, /* seek procedure */
      ucrSetOptionsProc, /* set options procedure */
      ucrGetOptionsProc, /* get options procedure */
      ucrWatchProc, /* watch procedure */
      ucrGetHandleProc /* get os handle procedure. */
};



/*
 * This is the generic fuction that implements all the methods of the
 * channel object.
 */
static int ucr_proc(ClientData cd, Tcl_Interp*interp, int argc, char*argv[])
{
      Tcl_Channel ch = (Tcl_Channel)cd;
      struct ucr_channel_info*info =
	    (struct ucr_channel_info*)Tcl_GetChannelInstanceData(ch);

      if (argc < 2)
	    return TCL_OK;

      if (strcmp(argv[1],"sync") == 0) {
	    ioctl(info->fd, UCR_SYNC, 0);
	    return TCL_OK;
      }

      if (strcmp(argv[1],"readable") == 0) {

	    if (info->read_cmd) {
		  free(info->read_cmd);
		  info->read_cmd = 0;
	    }

	    if (argc >= 3) {
		  info->read_cmd = malloc(strlen(argv[2]) + 1);
		  strcpy(info->read_cmd, argv[2]);
		  Tcl_CreateChannelHandler(ch, TCL_READABLE,
					   ucr_read_proc, (ClientData)info);

	    } else {
		  Tcl_DeleteChannelHandler(ch, ucr_read_proc,
					   (ClientData)info);
	    }

	    return TCL_OK;
      }

      if (strcmp(argv[1],"send_file") == 0) {

	    char buf[4096];
	    int idx;
	    for (idx = 2 ;  idx < argc ;  idx += 1) {
		  Tcl_Channel src = Tcl_OpenFileChannel(interp,
							argv[idx],
							"r", 0);
		  if (src == 0)
		      return TCL_ERROR;

		  Tcl_SetChannelOption(interp, src, "-translation", "binary");

		  while (!Tcl_Eof(src)) {
			int rc = Tcl_Read(src, buf, sizeof buf);
			if (rc < 0) {
			      Tcl_AppendResult(interp, "read error in ",
					       argv[idx], 0);
			      Tcl_Close(interp, src);
			      return TCL_ERROR;
			}
			if (rc > 0)
			      Tcl_Write(ch, buf, rc);
		  }
		  Tcl_Close(interp, src);
		  Tcl_Flush(ch);
	    }

	    return TCL_OK;
      }

      if (strcmp(argv[1],"send_var") == 0) {

	    int idx;
	    for (idx = 2 ;  idx < argc ;  idx += 1) {
		  Tcl_Obj*var = Tcl_NewStringObj(argv[1], -1);
		  Tcl_Obj*obj = Tcl_ObjGetVar2(interp, var, 0,
					       TCL_PARSE_PART1);
		  int len;
		  char*buf = Tcl_GetStringFromObj(obj, &len);

		  Tcl_Write(ch, buf, len);
		  TclFreeObj(var);
	    }

	    Tcl_Flush(ch);
	    return TCL_OK;
      }

      if (strcmp(argv[1],"test") == 0) {
	    ioctl(info->fd, UCR_TEST, 0);
	    return TCL_OK;
      }

      Tcl_SetResult(interp, "invalid subcommand", TCL_STATIC);
      return TCL_ERROR;
}


/*
 * The ucr_del_proc and ucr_open methods arranged for the uCR channel
 * objects to be created/destroyed in the tcl world.
 */

static void ucr_del_proc(ClientData cd)
{
      unsigned idx;
      Tcl_Channel ch = (Tcl_Channel)cd;
      struct ucr_channel_info*info =
	    (struct ucr_channel_info*)Tcl_GetChannelInstanceData(ch);

      struct ucr_frame_info*finfo;

	/* Close any frames associated with this channel. The
	   DeleteCommand will take care of deleting the finfo
	   structures by calling frame_del. */

      while (info->frames) {
	    Tcl_Interp*interp;
	    Tcl_Command cmd;
	    struct ucr_frame_info*finfo = info->frames;
	    info->frames = finfo->next;

	    interp = finfo->interp;
	    cmd = finfo->cmd;
	    Tcl_DeleteCommand(interp, Tcl_GetCommandName(interp,cmd));
      }

      Tcl_DeleteChannelHandler(ch, ucr_read_proc, (ClientData)info);
      Tcl_UnregisterChannel(info->interp, ch);

	/* Do not free the info structure here, allow the
	   UnregisterChannel to call close that in turn closes the
	   object. And _that_ is how the item gets deleted. */
}

static int ucr_open(ClientData cd, Tcl_Interp*interp, int argc, char*argv[])
{
      int rc;
      struct ucr_channel_info*info;
      int new_flag;

      if (argc < 4) {
	    Tcl_SetResult(interp, "usage: ucr_open <name> <path> <chan>",
			  TCL_STATIC);
	    return TCL_ERROR;
      }

      info = calloc(1, sizeof*info);
      assert(info);

      info->interp = interp;
      info->frames = 0;
      info->chan = strtoul(argv[3], 0, 0);
      info->self = Tcl_CreateHashEntry(&ucr_item_table, argv[1], &new_flag);
      if (! new_flag) {
	    Tcl_SetResult(interp, "name is busy.", TCL_STATIC);
	    free(info);
	    return TCL_ERROR;
      }

      Tcl_SetHashValue(info->self, info);


      rc = ucr_open_channel(info, argv[2], interp);
      if (rc != TCL_OK) {
	    Tcl_DeleteHashEntry(info->self);
	    free(info);
	    return TCL_ERROR;
      }

      info->block_mode = TCL_MODE_BLOCKING;
      info->flags = 0;
      info->tc = Tcl_CreateChannel(&ucr_channel_type, argv[1],
				   (ClientData)info,
				   TCL_READABLE|TCL_WRITABLE);
      Tcl_SetChannelOption(interp, info->tc, "-translation", "binary");
      Tcl_RegisterChannel(interp, info->tc);
      Tcl_CreateCommand(interp, argv[1], ucr_proc, (ClientData)info->tc,
			ucr_del_proc);

      return TCL_OK;
}



/*
 * The following is support for the uCR command device. This is not a
 * channel, but special access to the target for that allows certain
 * control operations to be performed.
 */
struct ucr_command_info {
      FILEDES fd;
};

static int ucrx_proc(ClientData cd, Tcl_Interp*interp, int argc, char*argv[])
{
      struct ucr_command_info*cx = (struct ucr_command_info*)cd;

      if (argc < 2)
	    return TCL_OK;

      if (strcmp(argv[1],"reset") == 0) {
	    int rc = ioctl(cx->fd, UCRX_RESET_BOARD, 0);
	    if (rc == -1) {
		  Tcl_SetResult(interp, "error reseting board.", TCL_STATIC);
		  return TCL_ERROR;
	    }

	    return TCL_OK;
      }

      if (strcmp(argv[1],"run") == 0) {
	    int rc = ioctl(cx->fd, UCRX_RUN_PROGRAM, 0);
	    if (rc == -1) {
		  switch (errno) {
#if !defined(WIN32)
		      case EBUSY:
			Tcl_SetResult(interp, "busy", TCL_STATIC);
			break;
		      case EAGAIN:
			Tcl_SetResult(interp, "no program", TCL_STATIC);
			break;
#endif
		      default:
			Tcl_SetResult(interp, "error", TCL_STATIC);
			break;
		  }
		  return TCL_ERROR;
	    }
	    return TCL_OK;
      }

      if (strcmp(argv[1],"trace") == 0) {
	    unsigned long arg = 0;
	    if (argc < 3) {
		  char buf[12];
		  long rc = ioctl(cx->fd, UCRX_GET_TRACE, 0);
		  if (rc < 0) {
			Tcl_SetResult(interp, "error getting trace",
				      TCL_STATIC);
			return TCL_ERROR;
		  }
		  sprintf(buf, "0x%x", rc);
		  Tcl_SetResult(interp, buf, TCL_VOLATILE);
		  return TCL_OK;
	    }

	    arg = strtoul(argv[2],0,0);
	    ioctl(cx->fd, UCRX_SET_TRACE, arg);
	    return TCL_OK;
      }

      if (strcmp(argv[1],"diagnose") == 0) {
	    int rc;
	    char buf[16];
	    unsigned long arg = 0;

	    if (argc >= 3)
		  arg = strtoul(argv[2],0,0);

	    rc = ioctl(cx->fd, UCRX_DIAGNOSE, arg);
	    sprintf(buf, "%d", rc);
	    Tcl_SetResult(interp, buf, TCL_VOLATILE);
	    return TCL_OK;
      }

      Tcl_SetResult(interp, "invalid subcommand.", TCL_STATIC);
      return TCL_ERROR;
}

static void ucrx_del_proc(ClientData cd)
{
      struct ucr_command_info*cx = (struct ucr_command_info*)cd;
#if defined(WIN32)
      CloseHandle(cx->fd);
#else
      close(cx->fd);
#endif
      free(cx);
}

static int ucrx_open(ClientData cd, Tcl_Interp*interp, int argc, char*argv[])
{
      struct ucr_command_info*cx;
      char fmt_path [64];
#if defined(WIN32)
      HANDLE fd;
#else
      int fd;
#endif

      if (argc < 3) {
	    Tcl_SetResult(interp, "wrong # args.", TCL_STATIC);
	    return TCL_ERROR;
      }

#if defined(WIN32)
      sprintf(fmt_path, "\\\\.\\%s", argv[2]);
      fd = CreateFile(fmt_path, GENERIC_READ|GENERIC_WRITE,
		      FILE_SHARE_READ|FILE_SHARE_WRITE, 0,
		      OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
      if (fd == INVALID_HANDLE_VALUE) {
	    Tcl_AppendResult(interp, "unable to open ``", argv[2],
			     "''", 0);
	    return TCL_ERROR;
      }
#else
      sprintf(fmt_path, "/dev/%s", argv[2]);
      fd = open(fmt_path, O_RDWR, 0);
      if (fd == -1) {
	    Tcl_AppendResult(interp, "unable to open ", argv[2], 0);
	    return TCL_ERROR;
      }
#endif


      cx = calloc(1, sizeof (struct ucr_command_info));
      cx->fd = fd;

      Tcl_CreateCommand(interp, argv[1], ucrx_proc, (ClientData)cx,
			ucrx_del_proc);
      return TCL_OK;
}

int Ucr_Init(Tcl_Interp*interp)
{
      Tcl_CreateCommand(interp, "ucr_open", ucr_open, 0, 0);
      Tcl_CreateCommand(interp, "ucr_frame", ucr_frame, 0, 0);
      Tcl_CreateCommand(interp, "ucrx_open", ucrx_open, 0, 0);

      Tcl_InitHashTable(&ucr_item_table, TCL_STRING_KEYS);
      Tcl_InitHashTable(&ucr_frame_table, TCL_STRING_KEYS);
      return TCL_OK;
}

/*
 * $Log: Ucr_Init.c,v $
 * Revision 1.23  1998/07/11 21:29:04  steve
 *  Frames on multiple ISE boards.
 *
 * Revision 1.22  1998/07/11 18:58:07  steve
 *  Fix leaking channel names.
 *
 * Revision 1.21  1998/07/10 00:41:34  steve
 *  Add unix support for fileevent.
 *
 * Revision 1.20  1998/06/11 19:06:42  steve
 *  Return length of allocated frame.
 *
 * Revision 1.19  1998/06/10 04:09:12  steve
 *  redo the frame interface.
 *
 * Revision 1.18  1998/05/29 18:41:05  steve
 *  Support for new SYNC ioctl.
 *
 * Revision 1.17  1998/05/07 16:42:40  steve
 *  Clear the name when deleted.
 *
 * Revision 1.16  1998/05/05 23:53:37  steve
 *  Software access to teh frame buffer.
 *
 * Revision 1.15  1998/04/17 22:29:01  steve
 *  Windows NT translation binary mode.
 *
 * Revision 1.14  1998/04/15 15:50:00  steve
 *  Trace control.
 *
 * Revision 1.13  1998/04/07 22:11:25  steve
 *  Tracing support.
 *
 * Revision 1.12  1998/04/03 01:35:13  steve
 *  Add assertion support.
 */

