#!/usr/bin/wish

# --
# Copyright (c) 1997,1998 Picture Elements, Inc.
# Stephen Williams (steve@picturel.com)
#
# $Id: ise_board.tcl,v 1.12 1998/07/11 22:43:50 steve Exp $
#
# This command uses the ucr device driver to implement a control panel
# for the ISE board. It is intended to work with the bootprom program
# included with the ISE configuration of uCR.
#
# The RESET button resets the i960 on the ISE board, and causes the
# BOOTPROM to take over the CPU. It does not yet assure that there are
# no processes or device drivers holding the device, that is up to you
# for now.
#
#    This source code is free software; you can redistribute it
#    and/or modify it in source code form under the terms of the GNU
#    General Public License as published by the Free Software
#    Foundation; either version 2 of the License, or (at your option)
#    any later version. In order to redistribute the software in
#    binary form, you will need a Picture Elements Binary Software
#    License.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
#
#    You should also have recieved a copy of the Picture Elements
#    Binary Software License offer along with the source. This offer
#    allows you to obtain the right to redistribute the software in
#    binary (compiled) form. If you have not received it, contact
#    Picture Elements, Inc., 777 Panoramic Way, Berkeley, CA 94704.
#



if {$tcl_platform(platform) == "windows"} {
    load ucr
    set config_path "ise_board.cfg"
} else {
    load libucr.so
    set config_path "~/ise_board.config"
}

proc reopen_configuration {ise} {
    global board_mode

    if {[scan $ise "ise%d" num] != 1} return

    catch {rename ucrx {}}
    ucrx_open ucrx isex$num
    set board_mode unknown
}

set ucrpath ise0
wm title . $ucrpath
reopen_configuration $ucrpath

trace variable ucrpath w trace_ucrpath
proc trace_ucrpath {x x x} {
    global ucrpath
    reopen_configuration $ucrpath
    wm title . $ucrpath
}

# Get the current trace bit values from the board. They could have
# been set by a previous execution of the program, other programs, or
# on loading the driver.
set tmp [ucrx trace]
set traceb(chan)  [expr $tmp & 1]
set traceb(frame) [expr $tmp & 2]
set traceb(proto) [expr $tmp & 4]
set traceb(ucrx)  [expr $tmp & 8]
unset tmp

# Read the config file for all the possible downloadable programs.
set config_file [open $config_path r]
set apps {}

while {[gets $config_file line] > 0} {
    lappend apps $line
}

# --
# This procedure uses the bridge unit to perform a hard reset of the
# i960 core and its local bus. This causes it to restart the bootprom
# image, and allow me to download programs into it.

proc reset_to_bootprom {} {
    global board_mode
    if {[catch {ucrx reset}]} {
	popup {Cannot reset the ISE board. It is probably busy.}
    } else {
	set board_mode reset
    }
}

# --
# This procedure uses the BIST flag to release the BOOTPROM to starting
# the application that has been loaded. If the BIST Capable bit is not
# set, then the program has not been successfully loaded and the start
# cannot proceed.

proc release_program {} {
    global board_mode

    if {[catch {ucrx run} msg]} {
	switch $msg {
	    busy {popup {Cannot start the application on the ISE board,\
		    because it is busy. Check that no processes have\
		    the board opened, and try again.}}

	    {no program} {popup {The application is not loaded \
		    (BIST Capable bit not set.) Load a linked program \
		    into the ISE device and try again.}}

	    default {popup {Error running program}}
	}
    } else {
	set board_mode running
    }
}

# --
# This procedure sends the named file down to the bootprom running on
# the target board.

proc download_program {file} {
    global ucrpath board_mode

    if {[string compare $board_mode reset] != 0} {
	popup {Board must be reset before downloading programs.}
	return
    }

    ucr_open tmp $ucrpath 0
    tmp send_file $file
    tmp sync
    rename tmp {}
    set board_mode standby
}

label .title -text "ISE Board Control Panel"

menubutton .board -text "Select Board" -menu .board.menu
menu .board.menu -tearoff false
.board.menu add radiobutton -label ise0 -variable ucrpath
.board.menu add radiobutton -label ise1 -variable ucrpath
.board.menu add radiobutton -label ise2 -variable ucrpath
.board.menu add radiobutton -label ise3 -variable ucrpath
button .reset -text "Reset to Bootprom" -command reset_to_bootprom
button .start -text "Start Application" -command release_program

menubutton .app -text "Download" -menu .app.menu
menu .app.menu -tearoff false

set idx 0
foreach app $apps {
    set label [lindex $app 0]
    set appfile [lindex $app 1]
    .app.menu add command -label $label -command "download_program $appfile"
    incr idx
}

frame .trace
checkbutton .trace.chan  -text Channels \
	-offvalue 0 -onvalue 1 -variable traceb(chan) -anchor w
checkbutton .trace.frame -text Frames \
	-offvalue 0 -onvalue 2 -variable traceb(frame) -anchor w
checkbutton .trace.proto -text Protocol \
	-offvalue 0 -onvalue 4 -variable traceb(proto) -anchor w
checkbutton .trace.ucrx  -text UCRX \
	-offvalue 0 -onvalue 8 -variable traceb(ucrx) -anchor w
pack .trace.chan .trace.frame .trace.proto .trace.ucrx \
	-side top -anchor w -expand true -fill x

trace variable traceb w traceb_command

proc traceb_command {name1 name2 op} {
    global traceb
    set mask 0
    incr mask $traceb(chan)
    incr mask $traceb(frame)
    incr mask $traceb(proto)
    incr mask $traceb(ucrx)
    ucrx trace $mask
}

button .diag0 -text "Diagnose 0" -command {ucrx diagnose 0}
button .diag1 -text "Diagnose 1" -command {ucrx diagnose 1}

pack .title .board .reset .start .app .trace .diag0 .diag1 -expand true -fill x

# --
# Build up a dialog box for displaying interesting messages. The only
# interactions applicable are simple acknowledgements of messages.

set message_box ""

toplevel .db

frame   .db.frame
label   .db.frame.title \
	-width 32 \
	-text "ISE Control Panel Message"
message .db.frame.message \
	-justify left \
	-relief groove \
	-aspect 500 \
	-bd 2 \
	-textvariable message_box
button  .db.frame.dismiss \
	-text "Dismiss" \
	-command {wm withdraw .db ; grab release .db}

pack .db.frame.title
pack .db.frame.message -expand 1 -fill x
pack .db.frame.dismiss
pack .db.frame

wm transient .db .
wm withdraw  .db

proc popup {message} {
    global message_box
    set message_box "$message"

    # place the dialog box over the control panel, offset just so.

    scan [wm geometry .] "%dx%d+%d+%d" w h x y
    set x [expr "$x + $w/3"]
    set y [expr "$y + $h/3"]
    wm geometry .db "+$x+$y"

    grab set .db
    wm deiconify .db
}

# $Log: ise_board.tcl,v $
# Revision 1.12  1998/07/11 22:43:50  steve
#  Device menu.
#
# Revision 1.11  1998/05/29 18:41:05  steve
#  Support for new SYNC ioctl.
#
# Revision 1.10  1998/05/26 16:16:35  steve
#  New channel protocol
#
# Revision 1.9  1998/04/25 05:03:00  steve
#  Handle displaying many programs.
#
# Revision 1.8  1998/04/17 22:27:31  steve
#  Properly manage the trace bits.
#
# Revision 1.7  1998/04/15 15:50:00  steve
#  Trace control.
#
# Revision 1.6  1998/04/07 22:11:25  steve
#  Tracing support.
#
# Revision 1.5  1998/04/03 01:50:37  steve
#  Add diagnose button.
#
# Revision 1.4  1998/04/03 01:30:01  steve
#  Support app downloading.
#
# Revision 1.3  1998/04/01 01:27:58  steve
#  Change device names to be consistent
#
# Revision 1.2  1997/12/12 04:29:38  steve
#  Install target typos.
#
# Revision 1.1  1997/12/12 01:01:34  steve
#  Add the tclif interface to ucr channels.
#
