/*
 * Copyright (c) 1998 Picture Elements, Inc.
 *    Stephen Williams (steve@picturel.com)
 *
 *    This source code is free software; you can redistribute it
 *    and/or modify it in source code form under the terms of the GNU
 *    General Public License as published by the Free Software
 *    Foundation; either version 2 of the License, or (at your option)
 *    any later version. In order to redistribute the software in
 *    binary form, you will need a Picture Elements Binary Software
 *    License.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *  ---
 *    You should also have recieved a copy of the Picture Elements
 *    Binary Software License offer along with the source. This offer
 *    allows you to obtain the right to redistribute the software in
 *    binary (compiled) form. If you have not received it, contact
 *    Picture Elements, Inc., 777 Panoramic Way, Berkeley, CA 94704.
 */
#if !defined(WINNT)
#ident "$Id: linux.c,v 1.9 1998/07/11 21:15:29 steve Exp $"
#endif

# include  <tcl.h>
# include  "priv.h"
# include  <errno.h>
# include  <ucrif.h>
# include  <assert.h>
# include  <unistd.h>
# include  <sys/mman.h>
# include  <string.h>

int ucrInputProc(ClientData cd, char*buf, int cnt, int*err)
{
      struct ucr_channel_info*info = (struct ucr_channel_info*)cd;
      int rc;

      if (info->block_mode == TCL_MODE_BLOCKING) {
	    fd_set readfds;
	    FD_ZERO(&readfds);
	    FD_SET(info->fd, &readfds);
	    select(info->fd+1, &readfds, 0, 0, 0);
      }

      rc = read(info->fd, buf, cnt);

      if (rc < 0)
	    *err = errno;

      return rc;
}

int ucrOutputProc(ClientData cd, char*buf, int cnt, int*err)
{
      struct ucr_channel_info*info = (struct ucr_channel_info*)cd;

      write(info->fd, buf, cnt);
      ioctl(info->fd, UCR_FLUSH, 0);
      *err = 0;

      return cnt;
}

static void ucr_file_proc(ClientData cd, int mask)
{
      struct ucr_channel_info*info = (struct ucr_channel_info*)cd;
      Tcl_NotifyChannel(info->tc, mask);
}

void ucrWatchProc(ClientData cd, int mask)
{
      struct ucr_channel_info*info = (struct ucr_channel_info*)cd;

      mask &= TCL_READABLE;
      if (mask) {
	    Tcl_CreateFileHandler(info->fd, mask, ucr_file_proc,
				  (ClientData)info);
      } else {
	    Tcl_DeleteFileHandler(info->fd);
      }
}

static unsigned frame_alloc_map = 0;

int ucr_manage_frame(Tcl_Interp*interp, struct ucr_frame_info*finfo,
		     unsigned long frame_size)
{
      int rc;
      unsigned long frame_id = finfo->num<<28;

      if (finfo->num >= 16) {
	    Tcl_SetResult(interp, "only frame 0-15 is supported.", TCL_STATIC);
	    return TCL_ERROR;
      }

      if (frame_size == 0) {
	    if (finfo->frame_base) {
		  munmap(finfo->frame_base, finfo->frame_size);
		  ioctl(finfo->cinfo->fd, UCR_FREE_FRAME, frame_id);
		  finfo->frame_base = 0;
		  finfo->frame_size = 0;
		  frame_alloc_map &= ~ (1<<finfo->num);
	    }
	    return TCL_OK;
      }

      assert( interp );

      if (frame_alloc_map & (1<<finfo->num)) {
	    Tcl_SetResult(interp, "frame is busy.", TCL_STATIC);
	    return TCL_ERROR;
      }

      frame_alloc_map |= 1<<finfo->num;
      rc = ioctl(finfo->cinfo->fd, UCR_MAKE_FRAME, frame_size|frame_id);
      if (rc == -1) {
	    Tcl_SetResult(interp, "Error making frame.", TCL_STATIC);
	    return TCL_ERROR;
      }

      finfo->frame_size = rc & 0x0fffffff;

      finfo->frame_base = mmap(0, frame_size, PROT_READ|PROT_WRITE,
			       MAP_SHARED, finfo->cinfo->fd, frame_id);

      return TCL_OK;
}

int ucr_open_channel(struct ucr_channel_info*info, const char*path,
		     Tcl_Interp*interp)
{
      int rc;
      char fmt_path[64];

      sprintf(fmt_path, "/dev/%s", path);
      info->fd = open(fmt_path, O_RDWR|O_NONBLOCK, 0);
      if (info->fd == -1) {
	    Tcl_AppendResult(interp, "unable to open ``", path, "''", 0);
	    return TCL_ERROR;
      }

      rc = ioctl(info->fd, UCR_CHANNEL, info->chan);
      if (rc == -1) {
	    Tcl_SetResult(interp, "unable to set channel number.", TCL_STATIC);
	    close(info->fd);
	    return TCL_ERROR;
      }

      return TCL_OK;
}

/*
 * $Log: linux.c,v $
 * Revision 1.9  1998/07/11 21:15:29  steve
 *  Clean up warnings.
 *
 * Revision 1.8  1998/07/10 00:41:34  steve
 *  Add unix support for fileevent.
 *
 * Revision 1.7  1998/06/18 01:11:48  steve
 *  Handle multiple frames.
 *
 * Revision 1.6  1998/06/10 04:09:12  steve
 *  redo the frame interface.
 *
 * Revision 1.5  1998/06/06 05:40:04  steve
 *  Linux/alpha 2.1.100 port.
 *
 * Revision 1.4  1998/04/01 01:27:58  steve
 *  Change device names to be consistent
 *
 * Revision 1.3  1998/03/23 19:56:09  steve
 *  Support for frames under NT,
 *  and non-blocking I/O under tcl/nt
 *
 * Revision 1.2  1998/03/12 23:00:25  steve
 *  Add frame commands.
 *
 * Revision 1.1  1998/03/12 21:32:28  steve
 *  divide the NT and linux implementataion up.
 *
 */

